###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageDB

.. module: StorageUsageDB

:synopsis: StorageUsageDB class is a front-end to the Storage Usage Database.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import six

import os
# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB

__RCSID__ = "$Id$"


def _standardDirectory(dirPath):
  """ This adds a / at the end of the path if not there """
  return os.path.join(dirPath, '')

#############################################################################


class StorageUsageDB(DB):
  """ .. class:: StorageUsageDB
  """

  def __init__(self):
    """Standard Constructor."""
    DB.__init__(self, 'StorageUsageDB', 'DataManagement/StorageUsageDB')
    self.__initializeDB()

  def __initializeDB(self):
    """Create the tables."""
    result = self._query("show tables")
    if not result['OK']:
      return result
    tablesInDB = [t[0] for t in result['Value']]
    tablesToCreate = {}
    self.__tablesDesc = {}
    self.__tablesDesc['su_Directory'] = {'Fields': {'DID': 'INTEGER UNSIGNED AUTO_INCREMENT NOT NULL',
                                                    'Path': 'VARCHAR(255) NOT NULL',
                                                    'Files': 'INTEGER UNSIGNED NOT NULL',
                                                    'Size': 'BIGINT UNSIGNED NOT NULL',
                                                    },
                                         'PrimaryKey': 'Path',
                                         'Indexes': {'id': ['dID']}
                                         }

    self.__tablesDesc['su_SEUsage'] = {'Fields': {'DID': 'INTEGER UNSIGNED NOT NULL',
                                                  'SEName': 'VARCHAR(32) NOT NULL',
                                                  'Files': 'INTEGER UNSIGNED NOT NULL',
                                                  'Size': 'BIGINT UNSIGNED NOT NULL',
                                                  'Updated': 'DATETIME NOT NULL'
                                                  },
                                       'PrimaryKey': ['DID', 'SEName'],
                                       'Indexes': {'SE': ['SEName']},
                                       }

    self.__tablesDesc['se_Usage'] = {'Fields': {'DID': 'INTEGER UNSIGNED NOT NULL',
                                                'SEName': 'VARCHAR(32) NOT NULL',
                                                'Files': 'INTEGER UNSIGNED NOT NULL',
                                                'Size': 'BIGINT UNSIGNED NOT NULL',
                                                'Updated': 'DATETIME NOT NULL'
                                                },
                                     'PrimaryKey': ['DID', 'SEName'],
                                     }
    self.__tablesDesc['se_STSummary'] = {'Fields': {'Site': 'VARCHAR(32) NOT NULL',
                                                    'SpaceToken': 'VARCHAR(32) NOT NULL',
                                                    'TotalSize': 'BIGINT UNSIGNED NOT NULL',
                                                    'TotalFiles': 'INTEGER UNSIGNED NOT NULL',
                                                    'Updated': 'DATETIME NOT NULL'
                                                    },
                                         'PrimaryKey': ['Site', 'SpaceToken'],
                                         }
    self.__tablesDesc['problematicDirs'] = {'Fields': {'DID': 'INTEGER UNSIGNED AUTO_INCREMENT NOT NULL',
                                                       'Path': 'VARCHAR(255) NOT NULL',
                                                       'Site': 'VARCHAR(32) NOT NULL',
                                                       'SpaceToken': 'VARCHAR(32) NOT NULL',
                                                       'ReplicaType': 'VARCHAR(32) NOT NULL',
                                                       'SEFiles': 'INTEGER UNSIGNED NOT NULL',
                                                       'LFCFiles': 'INTEGER UNSIGNED NOT NULL',
                                                       'SESize': 'BIGINT UNSIGNED NOT NULL',
                                                       'LFCSize': 'BIGINT UNSIGNED NOT NULL',
                                                       'Problem': 'VARCHAR(255) NOT NULL',
                                                       'Updated': 'DATETIME NOT NULL'
                                                       },
                                            'PrimaryKey': ['DID'],
                                            }
    self.__tablesDesc['Popularity'] = {'Fields': {'ID': 'INTEGER UNSIGNED AUTO_INCREMENT NOT NULL',
                                                  'Path': 'VARCHAR(255) NOT NULL',
                                                  'Site': 'VARCHAR(32) NOT NULL',
                                                  'Count': 'INTEGER UNSIGNED NOT NULL',
                                                  'InsertTime': 'DATETIME NOT NULL',
                                                  'Status': 'VARCHAR(32) NOT NULL'
                                                  },
                                       'PrimaryKey': ['ID'],
                                       'Indexes': {'InsertTime': ['InsertTime'], 'Status': ['Status']}
                                       }
    self.__tablesDesc['DirMetadata'] = {'Fields': {'DID': 'INTEGER UNSIGNED NOT NULL',
                                                   'ConfigName': 'VARCHAR(64) NOT NULL',
                                                   'ConfigVersion': 'VARCHAR(64) NOT NULL',
                                                   'Conditions': 'VARCHAR(128) NOT NULL',
                                                   'ProcessingPass': 'VARCHAR(255) NOT NULL',
                                                   'EventType': 'VARCHAR(255) NOT NULL',
                                                   'FileType': 'VARCHAR(64) NOT NULL',
                                                   'Production': 'VARCHAR(64) NOT NULL',
                                                   'Visibility': 'VARCHAR(4) NOT NULL'
                                                   },
                                        'PrimaryKey': ['DID'],
                                        }
    # These are templates for a possible popularity summary table
    # self.__tablesDesc['BKDatasets'] = { 'Fields': { 'DSID' : 'INTEGER UNSIGNED AUTO_INCREMENT NOT NULL',
    #                                                'DSPath':'VARCHAR(255) NOT NULL',
    #                                                'CreationTime':'DATETIME NOT NULL',
    #                                                'Archived':'BOOL'},
    #                                    'PrimaryKey' : ['DSID'],
    #                                    'Indexes' : { 'DSPath' : 'DSPath'}
    #                                  }
    # self.__tablesDesc['DatasetUsage'] = { 'Fields' : { 'DSID': 'INTEGER UNSIGNED NOT NULL',
    #                                                  'Week' : 'INTEGER UNSIGNED NOT NULL',
    #                                                  'Usage' : 'INTEGER UNSIGNED NOT NULL',
    #                                                  'Files' : 'INTEGER UNSIGNED NOT NULL',
    #                                                  'Size' : 'BIGINT UNSIGNED NOT NULL',
    #                                                  },
    #                                     'Indexes': { 'Week':'Week',
    #                                                 'DSID':'DSID'}
    #                                     }

    for tableName in self.__tablesDesc:
      if tableName not in tablesInDB:
        tablesToCreate[tableName] = self.__tablesDesc[tableName]
    return self._createTables(tablesToCreate)

  def __sqlDirList(self, dirList):
    """change a list of directories to an escape one."""
    return [self._escapeString(_standardDirectory(dirPath))['Value'] for dirPath in dirList]

  def __stripDirectory(self, dirPath):
    """Remove trailing / in directory names."""
    return self._escapeString(os.path.realpath(dirPath))['Value'][1:-1]

  ################
  # Bulk insertion
  ################

  def publishDirectories(self, directoryDict):
    """Inserts a group of directories with their usage."""

    if not directoryDict:
      return S_OK()
    self.log.debug("in publishDirectories: directoryDict is: \n %s" % str(directoryDict))
    result = self.__getIDs(directoryDict)
    if not result['OK']:
      return result
    dirIDs = result['Value']
    for path, pathInfo in directoryDict.items():
      path = _standardDirectory(path)
      try:
        files = int(pathInfo['Files'])
        size = int(pathInfo['TotalSize'])
      except ValueError as e:
        return S_ERROR("Values must be ints: %s" % repr(e))
      seUsage = pathInfo['SEUsage']
      sqlpath = self._escapeString(path)['Value']
      if path in dirIDs:
        sqlCmd = "UPDATE `su_Directory` SET Files=%d, Size=%d WHERE Path = %s" % (files, size, sqlpath)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot update directory", "%s: %s" % (path, result['Message']))
          return result
      else:
        sqlCmd = "INSERT INTO `su_Directory` (DID, Path, Files, Size) VALUES ( 0, %s, %d, %d )" % (sqlpath,
                                                                                                   files, size)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot insert directory", "%s: %s" % (path, result['Message']))
          continue
        dirIDs[path] = result['lastRowId']
      result = self.__updateSEUsage(dirIDs[path], seUsage, path)
      if not result['OK']:
        self.log.error("Cannot update SE usage", "%s:%s: %s" % (path, seUsage, result['Message']))
        return result
    return S_OK()

  ####
  # Insert/update to problematicDirs
  ####
  def publishToProblematicDirs(self, directoryDict):
    """Publish an entry into the problematic data directory."""
    for path, pathInfo in directoryDict.items():
      path = _standardDirectory(path)
      spaceToken = pathInfo['SpaceToken']
      problem = pathInfo['Problem']
      site = pathInfo['Site']
      replicaType = pathInfo['ReplicaType']
      try:
        seFiles = int(pathInfo['Files'])
        seSize = int(pathInfo['Size'])
        lfcFiles = int(pathInfo['LFCFiles'])
        lfcSize = int(pathInfo['LFCSize'])
      except ValueError:
        return S_ERROR("ERROR: Files and Size have to be integer!")
    # check if the tuple (path,Site,SpaceToken) already exists in the table
      sqlPath = self._escapeString(path)['Value']
      sqlSpaceToken = self._escapeString(spaceToken)['Value']
      sqlProblem = self._escapeString(problem)['Value']
      sqlSite = self._escapeString(site)['Value']
      sqlReplicaType = self._escapeString(replicaType)['Value']
      sqlCmd = "SELECT DID FROM problematicDirs WHERE Path=%s AND Site=%s AND SpaceToken=%s" % (sqlPath,
                                                                                                sqlSite,
                                                                                                sqlSpaceToken)
      self.log.verbose("in publishToProblematicDirs, sqlCmd: %s" % sqlCmd)
      result = self._query(sqlCmd)
      if not result['OK']:
        self.log.error("Failed to query problematicDirs")
        return result
      did = result['Value']
      if did:
        # there is an entry for (path, Site, SpaceToken), make an update of the row
        sqlCmd = "UPDATE `problematicDirs` SET SEFiles=%d, SESize=%d, LFCFiles=%d, " \
            "LFCSize=%d, Problem=%s, ReplicaType=%s, Updated=UTC_TIMESTAMP() WHERE " \
            "Path = %s and Site = %s and SpaceToken=%s" % (seFiles, seSize, lfcFiles, lfcSize, sqlProblem,
                                                           sqlReplicaType, sqlPath, sqlSite, sqlSpaceToken)
        self.log.verbose("in publishToProblematicDirs, sqlCmd: %s" % sqlCmd)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot update row (%s, %s) in problematicDirs" % (path, site))
          return result
      else:
        # entry is not there, make an insert of a new row
        sqlCmd = "INSERT INTO problematicDirs (Path, Site, SpaceToken, SEFiles, SESize, " \
            "LFCFiles, LFCSize, Problem, ReplicaType, Updated) VALUES " \
            "( %s, %s, %s, %d, %d, %d, %d, %s, %s, UTC_TIMESTAMP())" % (sqlPath, sqlSite, sqlSpaceToken,
                                                                        seFiles, seSize, lfcFiles,
                                                                        lfcSize, sqlProblem, sqlReplicaType)
        self.log.verbose("in publishToProblematicDirs, sqlCmd: %s" % sqlCmd)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot insert row (%s, %s) in problematicDirs" % (path, site))
          return result
    return S_OK()
  # ##
  # Insert/Update to se_Usage table
  # ##

  def publishToSEReplicas(self, directoryDict):
    """Publish an entry to se_Usage table."""
    for path, pathInfo in directoryDict.items():
      path = _standardDirectory(path)
      seName = pathInfo['SEName']
      try:
        files = int(pathInfo['Files'])
        size = int(pathInfo['Size'])
      except ValueError:
        return S_ERROR("ERROR: Files and Size have to be integer!")
      # check if the tuple (path,SE) already exists in the table
      sqlPath = self._escapeString(path)['Value']
      sqlSeName = self._escapeString(seName)['Value']
      sqlCmd = "SELECT d.DID FROM su_Directory AS d, se_Usage AS u WHERE " \
          "d.DID=u.DID AND d.Path=%s AND u.SEName=%s" % (sqlPath, sqlSeName)
      self.log.verbose("in publishToEReplias, sqlCmd: %s" % sqlCmd)
      result = self._query(sqlCmd)
      if not result['OK']:
        self.log.error("Failed to query se_Usage")
        return result
      sqlRes = result['Value']
      if sqlRes:
        try:
          did = int(sqlRes[0][0])
        except ValueError:
          return S_ERROR("ERROR: DID should be integer!")
        # there is an entry for (path, SEname), make an update of the row
        sqlCmd = "UPDATE `se_Usage` SET Files=%d, Size=%d, Updated=UTC_TIMESTAMP() WHERE " \
            "DID=%d AND SEName=%s" % (files, size, did, sqlSeName)
        self.log.verbose("in publishToEReplias, sqlCmd: %s" % sqlCmd)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot update row (%s, %s) in se_Usage" % (path, seName))
          return result
      else:
        # entry is not there, make an insert of a new row
        # get the DID corresponding to this Path, from su_Directory table:
        result = self.__getIDs(directoryDict)
        if not result['OK']:
          self.log.error("Cannot getIds for directory %s" % (path))
          return result
        for dirKey in result['Value']:
          try:
            did = int(result['Value'][dirKey])
          except ValueError:
            return S_ERROR("ERROR: DID should be integer!")
        sqlCmd = "INSERT INTO se_Usage (DID, SEName, Files, Size, Updated) VALUES " \
            " ( %d, %s, %d, %d, UTC_TIMESTAMP())" % (did, sqlSeName, files, size)
        self.log.verbose("in publishToEReplias, sqlCmd: %s" % sqlCmd)
        result = self._update(sqlCmd)
        if not result['OK']:
          self.log.error("Cannot insert row (%s, %s) in se_Usage" % (path, seName))
          return result
    return S_OK()

  def getIDs(self, dirList):
    """get IDs for list of directories."""
    return self.__getIDs(dirList)

  def __getIDs(self, dirList):
    """get IDs for list of directories."""
    dl = self.__sqlDirList(dirList)
    sqlCmd = "SELECT Path, DID FROM `su_Directory` WHERE Path in ( %s )" % ", ".join(dl)
    self.log.verbose("in __getIDs, query: %s" % sqlCmd)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK(dict(result['Value']))

  def __getIDsFromProblematicDirs(self, dirList):
    """Get the ID from the problematicDirs table, for a given directory
    {Path:seInfo}"""
    self.log.verbose("entry to be removed: %s" % dirList)
    dirPath, dirInfo = list(dirList.items())[0]
    dirPath = _standardDirectory(dirPath)
    spaceToken = dirInfo['SpaceToken']
    sqlPath = self._escapeString(dirPath)['Value']
    sqlSpaceToken = self._escapeString(spaceToken)['Value']
    site = dirInfo['Site']
    sqlSite = self._escapeString(site)['Value']
    sqlCmd = "SELECT Path, DID FROM problematicDirs WHERE Path=%s AND Site=%s and SpaceToken=%s" % (sqlPath,
                                                                                                    sqlSite,
                                                                                                    sqlSpaceToken)
    self.log.verbose("in __getIDsFromProblematicDirs, query: %s" % sqlCmd)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK(dict(result['Value']))

  def __getIDsFromSe_Usage(self, dirList):
    """Get the ID of the entry corresponding to the directory {path:seInfo}"""
    dirPath, dirInfo = list(dirList.items())[0]
    dirPath = _standardDirectory(dirPath)

    # take into account that SEName might not be available
    se = ''
    site = ''
    try:
      se = dirInfo['SEName']
    except KeyError:
      self.log.info("SEName attribute not available!")
    if not se:
      self.log.info("The SEName attribute is not available. Try to get the Site attribute..")
      try:
        site = dirInfo['Site']
      except KeyError:
        self.log.info("Site attribute not available!")
    if se:
      sqlSE = self._escapeString(se)['Value']
    else:
      sqlSE = ''
    if site:
      sqlSite = self._escapeString(site)['Value']
    else:
      sqlSite = ''
    sqlPath = self._escapeString(dirPath)['Value']
    if sqlSE:
      sqlCmd = "SELECT d.Path, d.DID FROM se_Usage r, su_Directory d WHERE d.DID=r.DID \
      AND d.Path=%s AND r.SEName=%s" % (sqlPath, sqlSE)
    elif sqlSite:
      sqlCmd = "SELECT d.Path, d.DID FROM se_Usage r, su_Directory d WHERE d.DID=r.DID \
      AND d.Path=%s AND r.SEName LIKE '%s%%'" % (sqlPath, sqlSite)
    else:
      self.log.error("Not enough information to formulate the query: either Site or SEName should be provided ")
      return S_ERROR()
    self.log.verbose("in __getIDsFromSe_Usage, query: %s" % sqlCmd)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK(dict(result['Value']))

  def __updateSEUsage(self, dirID, seUsage, path):
    """Update the SE usage table."""
    sqlCmd = "DELETE FROM `su_SEUsage` WHERE DID=%d" % dirID
    result = self._update(sqlCmd)
    if not result['OK']:
      return result
    if not seUsage:
      # This indicates there are files without replicas in the FC
      self.log.warn("Ooops. Dir has no SEUsage!", "ID %s, path %s" % (dirID, path))
      return S_OK()
    sqlVals = []
    # Insert data
    for seName in seUsage:
      try:
        size = seUsage[seName]['Size']
        files = seUsage[seName]['Files']
        # Protect against hiccoughs of the DFC that may return negative numbers
        if size < 0 or files < 0:
          self.log.error("SEUsage < 0 for %s @ %s: %d files, %d size" % (path, seName, files, size))
          continue
        # This line should be in the try otherwise the exception will never work
        sqlVals.append("( %d, %s, %d, %d, UTC_TIMESTAMP() )" % (dirID, self._escapeString(seName)['Value'],
                                                                size, files))
      except ValueError as e:
        return S_ERROR("Values must be ints: %s" % repr(e))
    # It may happen that no SE is OK
    if not sqlVals:
      return S_OK()
    sqlIn = "INSERT INTO `su_SEUsage` ( DID, SEname, Size, Files, Updated ) VALUES %s" % ", ".join(sqlVals)
    return self._update(sqlIn)

  #############
  # Remove dir
  #############

  def __getIDsLike(self, dirPath):
    dirPath = self.__stripDirectory(dirPath)
    sqlCmd = "SELECT Path, DID FROM `su_Directory` WHERE Path LIKE '%s/%%'" % dirPath
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK(dict(result['Value']))

  def removeDirectory(self, dirPath):
    result = self.__getIDsLike(dirPath)
    if not result['OK']:
      return result
    dirIDs = result['Value']
    if not dirIDs:
      return S_OK()
    sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
    for tableName in ('su_Directory', 'su_SEUsage'):
      sqlCmd = "DELETE FROM `%s` WHERE DID in ( %s )" % (tableName, sqlIDs)
      result = self._update(sqlCmd)
      if not result['OK']:
        return result
    return S_OK()

  def removeDirFromSe_Usage(self, dirDict):
    """Remove the entry corresponding to the tuple (path, SE) from the se_Usage
    table.

    This function is typically called when a directory is found to be a
    problematic directory, and before inserting it into the
    problematicDirs table, it is necessary to remove it from the
    se_Usage table, if it exists there. In general, one same directory
    can only exist into either the se_Usage or the problematicDirs
    table.
    """
    deletedDirs = 0
    result = self.__getIDsFromSe_Usage(dirDict)
    if not result['OK']:
      return result
    dirIDs = result['Value']
    if not dirIDs:
      return S_OK(deletedDirs)
    deletedDirs = len(dirIDs)
    if deletedDirs > 1:
      return S_ERROR("There should not be more than 1 match for a tuple (path,SE)!!")
    sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
    sqlCmd = "DELETE FROM se_Usage WHERE DID in ( %s )" % (sqlIDs)
    result = self._update(sqlCmd)
    if not result['OK']:
      return result
    return S_OK(deletedDirs)

  def removeDirFromProblematicDirs(self, dirDict):
    """Remove an entry from the problematicDirs table.

    This is typically used when a directory is found to be correctly
    registered in the FC and it was previously inserted in the
    problematic data table: before inserting the entry into the se_Usage
    table, it has to be removed from the problematicDirs
    """
    deletedDirs = 0
    result = self.__getIDsFromProblematicDirs(dirDict)
    if not result['OK']:
      return result
    self.log.verbose("sqlCmd result for __getIDsFromProblematicDirs is: %s" % result)
    dirIDs = result['Value']
    if not dirIDs:
      self.log.info("No directory to remove")
      return S_OK(deletedDirs)
    deletedDirs = len(dirIDs)
    if deletedDirs > 1:
      return S_ERROR("There should not be more than 1 match for a tuple (path,SE)!!")
    self.log.info("remove directories: %s" % dirIDs)
    sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
    sqlCmd = "DELETE FROM problematicDirs WHERE DID in ( %s )" % (sqlIDs)
    result = self._update(sqlCmd)
    if not result['OK']:
      return result
    self.log.verbose("sqlCmd result for DELETE is: %s" % result)
    return S_OK(deletedDirs)

  def removeAllFromProblematicDirs(self, site=False):
    """Remove all entries from the problematicDirs table, for a given site
    (optional)."""
    if site:
      sqlSite = self._escapeString(site)['Value']
      sqlCmd = "DELETE FROM problematicDirs WHERE Site=%s" % sqlSite
    else:
      sqlCmd = "DELETE FROM problematicDirs"
    self.log.verbose("in removeAllFromProblematicDirs, command: %s" % sqlCmd)
    result = self._update(sqlCmd)
    if not result['OK']:
      return result
    self.log.verbose("sqlCmd result for DELETE is: %s" % result)
    deletedDirs = result['Value']
    return S_OK(deletedDirs)

  ################
  # Clean outdated entries
  ################

  def purgeOutdatedEntries(self, rootDir=False, outdatedSeconds=86400, preserveDirsList=None):
    preserveDirsList = preserveDirsList if preserveDirsList else []
    try:
      outdatedSeconds = max(1, int(outdatedSeconds))
    except ValueError:
      return S_ERROR("Ooutdated seconds needs to be a number")

    sqlCond = ["TIMESTAMPDIFF( SECOND, su.Updated, UTC_TIMESTAMP() ) > %d" % outdatedSeconds]
    sqlTables = ["`su_SEUsage` as su"]
    sqlLimit = 1000
    if rootDir:
      rootDir = self.__stripDirectory(rootDir)
      sqlTables.append("`su_Directory` as d")
      sqlCond.append("d.Path LIKE '%s/%%'" % rootDir)
      sqlCond.append("d.DID = su.DID")
    if preserveDirsList:
      for ignoreDir in preserveDirsList:
        ignoreDir = self.__stripDirectory(ignoreDir)
        sqlCond.append("d.Path NOT LIKE '%s/%%'" % ignoreDir)
    sqlCmd = "SELECT DISTINCT su.DID FROM %s WHERE %s LIMIT %d" % (", ".join(sqlTables),
                                                                   " AND ".join(sqlCond),
                                                                   sqlLimit)
    cleaned = 0
    while True:
      result = self._query(sqlCmd)
      if not result['OK']:
        return result
      idList = [str(row[0]) for row in result['Value']]
      if not idList:
        break
      cleaned += len(idList)
      sqlIdList = ", ".join(idList)
      for tblName in ('su_Directory', 'su_SEUsage'):
        result = self._update("DELETE FROM `%s` WHERE DID IN ( %s )" % (tblName, sqlIdList))
        if not result['OK']:
          return result
      if len(idList) < sqlLimit:
        break
    return S_OK(cleaned)

  ###############
  #  Get Storage info
  ###############

  def getStorageElementSelection(self):
    """Retireve the possible selections available through the web-monitor."""
    sqlCmd = "SELECT DISTINCT SEName FROM `su_SEUsage` ORDER BY SEName"
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK([row[0] for row in result['Value']])

  def __getStorageCond(self, path, fileType, production, SEs):
    path = self.__stripDirectory(path)
    sqlCond = ["d.DID = su.DID", "d.Path LIKE '%s/%%'" % path]
    if fileType:
      fileType = self._escapeString(fileType)['Value'][1:-1]
      sqlCond.append("d.Path LIKE '%%/%s%%'" % fileType)
    if production:
      try:
        sqlCond.append("d.Path LIKE '%%/%08.d%%'" % int(production))
      except ValueError:
        return S_ERROR("production has to be a number")
    if SEs:
      sqlCond.append("su.SEName in ( %s )" % ", ".join([self._escapeString(SEName)['Value']
                                                        for SEName in SEs]))
    return sqlCond

  def __getStorageSummary(self, path, fileType=False, production=False, SEs=None, groupingField="su.SEName"):
    """Retrieves the storage summary for all of the known directories."""
    if SEs is None:
      SEs = []
    sqlCond = self.__getStorageCond(path, fileType, production, SEs)
    sqlFields = (groupingField, "SUM(su.Size)", "SUM(su.Files)")
    sqlCmd = "SELECT %s FROM `su_SEUsage` as su,`su_Directory` as d WHERE %s GROUP BY %s" % (", ".join(sqlFields),
                                                                                             " AND ".join(sqlCond),
                                                                                             groupingField)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    usageDict = {}
    for gf, size, files in result['Value']:
      usageDict[gf] = {'Size': int(size), 'Files': int(files)}
    return S_OK(usageDict)

  def getStorageSummary(self, path, fileType=False, production=False, SEs=[]):
    return self.__getStorageSummary(path, fileType, production, SEs, "su.SEName")

  def getStorageDirectorySummary(self, path, fileType=False, production=False, SEs=[]):
    return self.__getStorageSummary(path, fileType, production, SEs, "d.Path")

  def getStorageDirectories(self, path, fileType=False, production=False, SEs=[]):
    sqlCond = self.__getStorageCond(path, fileType, production, SEs)
    sqlCmd = "SELECT DISTINCT d.Path FROM `su_SEUsage` as su,`su_Directory` as d WHERE %s" % (" AND ".join(sqlCond))
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    return S_OK([row[0] for row in result['Value']])

  @staticmethod
  def __userExpression(fieldName="d.Path"):
    return "SUBSTRING_INDEX( SUBSTRING_INDEX( %s, '/', 5 ), '/', -1 )" % fieldName

  def getUserStorageUsage(self, userName=False):
    """Get the usage in the SEs."""
    sqlCond = ["d.DID = su.DID"]
    if userName:
      userName = self._escapeString(userName)['Value'][1:-1]
      sqlCond.append("d.Path LIKE '/lhcb/user/%s/%s/%%" % (userName[0], userName))
    else:
      sqlCond.append("d.Path LIKE '/lhcb/user/_/%'")
    sqlCmd = "SELECT %s, SUM( su.Size ) FROM `su_Directory` AS d, `su_SEUsage` AS su " \
        "WHERE %s GROUP BY %s" % (self.__userExpression(), " AND ".join(sqlCond), self.__userExpression())
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    userStorage = {}
    for row in result['Value']:
      userStorage[row[0]] = int(row[1])
    return S_OK(userStorage)

  def getUserSummaryPerSE(self, userName=False):
    sqlUser = self.__userExpression()
    sqlFields = (sqlUser, "su.SEName", "SUM(su.Size)", "SUM(su.Files)")
    sqlCond = ["d.DID = su.DID", "d.Path LIKE '/lhcb/user/_/%'"]
    if userName:
      sqlCond.append("%s = %s" % (sqlUser, self._escapeString(userName)['Value']))
    sqlGroup = (sqlUser, "su.SEName")
    sqlCmd = "SELECT % s FROM `su_Directory` AS d, `su_SEUsage` AS su WHERE %s GROUP BY %s" % (", ".join(sqlFields),
                                                                                               " AND ".join(sqlCond),
                                                                                               ", ".join(sqlGroup))
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    userData = {}
    for row in result['Value']:
      userName = row[0]
      seName = row[1]
      if userName not in userData:
        userData[userName] = {}
      userData[userName][seName] = {'Size': int(row[2]), 'Files': int(row[3])}
    return S_OK(userData)

  def getDirectorySummaryPerSE(self, directory):
    """Queries the DB and get a summary (total size and files) for the given
    directory."""
    sqlCmd = "SELECT su.SEName, SUM(su.Size), SUM(su.Files) FROM su_Directory AS d, su_SEUsage AS su " \
        " WHERE d.DID=su.DID AND d.Path LIKE '%s%%' GROUP BY su.SEName" % (directory)
    self.log.verbose("in getDirectorySummaryPerSE, sqlCmd: %s " % sqlCmd)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    data = {}
    for row in result['Value']:
      seName = row[0]
      if seName not in data:
        data[seName] = {}
      data[seName] = {'Size': int(row[1]), 'Files': int(row[2])}
    return S_OK(data)

  def publishTose_STSummary(self, site, spaceToken, totalSize, totalFiles, storageDumpLastUpdate):
    """Publish total size and total files extracted from the storage dumps to
    the se_STSummary."""
    try:
      sqlTotalSize = int(totalSize)
      sqlTotalFiles = int(totalFiles)
    except ValueError as e:
      return S_ERROR("Values must be ints: %s" % repr(e))
    sqlSpaceToken = self._escapeString(spaceToken)['Value']
    sqlSite = self._escapeString(site)['Value']
    sqlSDUpdate = self._escapeString(storageDumpLastUpdate)['Value']
    # check if there is already an entry for the tuple (site, spaceTok)
    res = self.__getSTSummary(site, spaceToken)
    self.log.verbose(" __getSTSummary returned: %s " % res)
    if not res['OK']:
      return S_ERROR(res)
    if not res['Value']:
      self.log.warn("Entry for site=%s, spaceToken=%s is not there => insert new entry " % (site, spaceToken))
      sqlCmd = "INSERT INTO `se_STSummary` (Site, SpaceToken, TotalSize, TotalFiles, Updated) VALUES " \
          "( %s, %s, %d, %d, %s)" % (sqlSite, sqlSpaceToken, sqlTotalSize, sqlTotalFiles, sqlSDUpdate)

    else:
      self.log.verbose("Entry for site=%s, spaceToken=%s is there => update entry " % (site, spaceToken))
      sqlCmd = "UPDATE `se_STSummary` SET TotalSize=%d, TotalFiles=%d, Updated=%s WHERE " \
          "Site=%s and SpaceToken=%s" % (sqlTotalSize, sqlTotalFiles, sqlSDUpdate, sqlSite, sqlSpaceToken)

    self.log.verbose("in publishTose_STSummary, sqlCmd: %s " % sqlCmd)
    result = self._update(sqlCmd)
    if not result['OK']:
      self.log.error("Cannot insert entry", "%s,%s: %s" % (sqlSite, sqlSpaceToken, result['Message']))
      return result
    return S_OK()

  def __getSTSummary(self, site, spaceToken=False):
    """Get total files and total size for the space token identified by the
    input arguments (site, space token) , if space token is not specified,
    return all entries relative to the site."""

    sqlSite = self._escapeString(site)['Value']
    sqlCond = ["Site=%s" % sqlSite]
    if spaceToken:
      sqlSpaceToken = self._escapeString(spaceToken)['Value']
      sqlCond.append("SpaceToken=%s" % sqlSpaceToken)

    sqlCmd = "SELECT Site, SpaceToken, TotalSize, TotalFiles, Updated FROM `se_STSummary` WHERE %s" % \
        " AND ".join(sqlCond)
    self.log.verbose("in __getSTSummary sqlCmd is: %s " % sqlCmd)
    result = self._query(sqlCmd)
    self.log.verbose(" __getSTSummary result: %s " % result)
    if not result['OK']:
      return S_ERROR(result)
    return S_OK(result['Value'])

  def getSTSummary(self, site, spaceToken=False):
    """Returns a summary of space usage for the given site, on the basis of the
    information provided by the SRM db dumps from sites."""
    return self.__getSTSummary(site, spaceToken)

  def removeSTSummary(self, site, spaceToken=False):
    """Remove from se_STSummary table the entry relative to the given site and
    space token (if specified).

    If no space token is specified, remove all entries relative to site
    """
    sqlSite = self._escapeString(site)['Value']
    sqlCond = ["Site=%s" % sqlSite]
    if spaceToken:
      sqlSpaceToken = self._escapeString(spaceToken)['Value']
      sqlCond.append("SpaceToken=%s" % sqlSpaceToken)

    sqlCmd = "DELETE FROM `se_STSummary` WHERE %s" % (" AND ".join(sqlCond))
    self.log.verbose("in __getSTSummary, sqlCmd: %s " % sqlCmd)
    result = self._update(sqlCmd)
    self.log.verbose("in __getSTSummary result: %s " % result)
    if not result['OK']:
      return S_ERROR(result)
    return S_OK(result['Value'])

  def getProblematicDirsSummary(self, site, problem=False):
    """Get a summary of problematic directories for a given site and given
    problem (optional)"""
    sqlSite = self._escapeString(site)['Value']
    sqlCond = ["Site=%s" % sqlSite]
    if problem:
      sqlProblem = self._escapeString(problem)['Value']
      sqlCond.append("Problem=%s" % sqlProblem)

    sqlCmd = "SELECT Site, SpaceToken, LFCFiles, SEFiles, Path, Problem, ReplicaType FROM `problematicDirs` " \
        "WHERE %s" % (" AND ".join(sqlCond))
    self.log.verbose("in getProblematicDirsSummary sqlCmd: %s " % sqlCmd)
    result = self._query(sqlCmd)
    self.log.verbose("in getProblematicDirsSummary result: %s " % result)
    if not result['OK']:
      return S_ERROR(result['Message'])
    return S_OK(result['Value'])

  def getRunSummaryPerSE(self, run):
    """Queries the DB and get a summary (total size and files) per SE  for the
    given run. It assumes that the path in the LFC where the run's file are
    stored is like:

    /lhcb/data/[YEAR]/RAW/[STREAM]/[PARTITION]/[ACTIVITY]/[RUNNO]/
    """
    # try and implement bulk query
    # check the type of run
    data = {}
    if isinstance(run, list):
      for thisRun in run:
        sqlCmd = "SELECT su.SEName, SUM(su.Size), SUM(su.Files)  FROM su_Directory AS d, su_SEUsage AS su WHERE " \
            "d.DID=su.DID AND d.Path LIKE '/lhcb/data/%%/RAW/%%/%%/%%/%d/' GROUP BY su.SEName" % (thisRun)
        self.log.verbose("in getRunSummaryPerSE, sqlCmd: %s " % sqlCmd)
        result = self._query(sqlCmd)
        if not result['OK']:
          return S_ERROR(result)
        data[thisRun] = {}
        for row in result['Value']:
          seName = row[0]
          data[thisRun][seName] = {'Size': int(row[1]), 'Files': int(row[2])}
    else:
      sqlCmd = "SELECT su.SEName, SUM(su.Size), SUM(su.Files)  FROM su_Directory AS d, su_SEUsage AS su WHERE " \
          "d.DID=su.DID and d.Path LIKE '/lhcb/data/%%/RAW/%%/%%/%%/%d/' GROUP BY su.SEName" % (run)
      self.log.verbose("in getRunSummaryPerSE, sqlCmd: %s " % sqlCmd)
      result = self._query(sqlCmd)
      if not result['OK']:
        return S_ERROR(result)
      for row in result['Value']:
        seName = row[0]
        data[seName] = {'Size': int(row[1]), 'Files': int(row[2])}

    return S_OK(data)

  def __getAllReplicasInFC(self, path):
    """Queries the su_seUsage table to get all the entries relative to a given
    path registered in the FC.

    Returns for every replica the SE, the update, the files and the
    total size
    """
    path = _standardDirectory(path)
    sqlCmd = "SELECT DID FROM su_Directory where Path like '%s%%'" % (path)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    if not result['Value']:  # no replica found
      return S_OK(result)

    for row in result['Value']:
      did = row[0]
      # warning! this has to be corrected! there is no need to make a loop! the query should return only one row
      # add a check for that. And return an error if it return more than one row

    sqlCmd = "SELECT Files, Updated, SEName, Size FROM su_SEUsage WHERE DID=%s" % (did)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    replicasData = {}
    replicasData[path] = {}
    for row in result['Value']:
      seName = row[2]
      if seName in replicasData[path]:
        return S_ERROR("There cannot be two replicas on the same SE!")
      replicasData[path][seName] = {}
      replicasData[path][seName]['Files'] = row[0]
      replicasData[path][seName]['Updated'] = row[1]
      replicasData[path][seName]['Size'] = row[3]

    return S_OK(replicasData)

  def getAllReplicasInFC(self, path):
    return self.__getAllReplicasInFC(path)

  ######
  # Catalog queries
  ######

  def __getCatalogCond(self, path, fileType, production):
    path = self.__stripDirectory(path)
    sqlCond = ["Path LIKE '%s/%%'" % path]
    if fileType:
      fileType = self._escapeString(fileType)['Value'][1:-1]
      sqlCond.append("Path LIKE '%%/%s%%'" % fileType)
    if production:
      try:
        sqlCond.append("Path LIKE '%%/%08.d%%'" % int(production))
      except ValueError:
        return S_ERROR("production has to be a number")
    return sqlCond

  def __getSummary(self, path, fileType=False, production=False, groupingField="Path"):
    """Retrieves the storage summary for all of the known directories."""
    sqlCond = self.__getCatalogCond(path, fileType, production)
    sqlFields = (groupingField, "SUM(Size)", "SUM(Files)")
    sqlCmd = "SELECT %s FROM `su_Directory` WHERE %s GROUP BY %s" % (", ".join(sqlFields),
                                                                     " AND ".join(sqlCond),
                                                                     groupingField)
    result = self._query(sqlCmd)
    if not result['OK']:
      return result
    usageDict = {}
    for gf, size, files in result['Value']:
      usageDict[gf] = {'Size': int(size), 'Files': int(files)}
    return S_OK(usageDict)

  def getSummary(self, path, fileType=False, production=False):
    return self.__getSummary(path, fileType, production, "Path")

  def getUserSummary(self, userName=False):
    if userName:
      userName = self._escapeString(userName)['Value'][1:-1]
      path = "/lhcb/user/%s/%s" % (userName[0], userName)
    else:
      path = "/lhcb/user/_/%"
    return self.__getSummary(path, groupingField=self.__userExpression("Path"))

  ######
  # methods to interact with Popularity table
  ######

  def sendDataUsageReport(self, site, directoryDict, status='New'):
    """Add a new trace in the Popularity table."""
    self.log.verbose("in addPopCount: dirDict: %s site: %s" % (directoryDict, site))
    if not directoryDict:
      return S_OK()
    sqlSite = self._escapeString(site)['Value']
    insertedEntries = 0
    for di, count in directoryDict.items():
      di = _standardDirectory(di)
      sqlPath = self._escapeString(di)['Value']
      sqlStatus = self._escapeString(status)['Value']
      if not isinstance(count, int):
        self.log.warn("in sendDataUsageReport: type is not correct %s" % count)
        continue
      sqlCmd = "INSERT INTO `Popularity` ( Path, Site, Count, Status, InsertTime) VALUES " \
          "( %s, %s, %d, %s, UTC_TIMESTAMP() )" % (sqlPath, sqlSite, count, sqlStatus)
      self.log.verbose("in sendDataUsageReport, sqlCmd: %s " % sqlCmd)
      result = self._update(sqlCmd)
      if not result['OK']:
        self.log.error("Cannot insert entry: %s" % (result['Message']))
        continue
      insertedEntries += 1
    return S_OK(insertedEntries)

  def sendDataUsageReport_2(self, directoryDict):
    """Add a new trace in the Popularity table (new version which takes in
    input a dictionary per trace) Each trace corresponds to a directory, and
    the mandatory keys are the site and the count.

    Optional keys are the status and the creation time
    """
    self.log.verbose("in addPopCount: dirDict: %s" % (directoryDict))
    if not directoryDict:
      return S_OK()
    insertedEntries = 0
    for di in directoryDict:
      try:
        site = directoryDict[di]['site']
        count = directoryDict[di]['count']
      except KeyError:
        self.log.error("input directoryDict should specify site and count keys . %s " % str(directoryDict))
        return S_ERROR('wrong arguments format')
      # set default value for status
      status = directoryDict[di].get('status', 'New')

      di = _standardDirectory(di)
      sqlPath = self._escapeString(di)['Value']
      sqlStatus = self._escapeString(status)['Value']
      sqlSite = self._escapeString(site)['Value']
      if not isinstance(count, int):
        self.log.warn("in sendDataUsageReport: type is not correct %s" % count)
        continue
      # by default, insert the record with the current time, unless the input directoryDict specifies the creation time
      insertTime = directoryDict[di].get('creationTime')
      sqlInsertTime = self._escapeString(insertTime)['Value'] if insertTime else 'UTC_TIMESTAMP()'
      sqlCmd = "INSERT INTO `Popularity` ( Path, Site, Count, Status, InsertTime) VALUES " \
          "( %s, %s, %d, %s, %s )" % (sqlPath, sqlSite, count, sqlStatus, sqlInsertTime)
      self.log.verbose("in sendDataUsageReport_2, sqlCmd: %s " % sqlCmd)
      result = self._update(sqlCmd)
      if not result['OK']:
        self.log.error("Cannot insert entry: %s" % (result['Message']))
        continue
      insertedEntries += 1
    return S_OK(insertedEntries)

  def getDataUsageSummary(self, startTime, endTime, status='New'):
    """returns a summary of the counts for each tuple (Site,Path) in the given
    time interval."""
    if startTime > endTime:
      return S_OK()
    if not isinstance(startTime, six.string_types) or not isinstance(endTime, six.string_types):
      return S_ERROR('wrong arguments format')

    sqlStartTime = self._escapeString(startTime)['Value']
    sqlEndTime = self._escapeString(endTime)['Value']
    sqlStatus = self._escapeString(status)['Value']
    sqlCmd = "SELECT ID, Path, Site, Count, InsertTime from `Popularity` WHERE " \
        "Status = %s AND InsertTime > %s AND InsertTime < %s " % (sqlStatus, sqlStartTime, sqlEndTime)
    self.log.verbose("in getDataUsageSummary, sqlCmd: %s " % sqlCmd)
    return self._query(sqlCmd)

  def getDataUsageForDirectory(self, path):
    """returns all entries for a given path."""
    sqlCmd = "SELECT ID, Path, Site, Count, InsertTime from `Popularity` WHERE Path = %s" % \
        self._escapeString(path)['Value']
    return self._query(sqlCmd)

  def updatePopEntryStatus(self, idList, newStatus):
    """Update the status of the entry identified by the IDList, to the status
    specified by Newstatus."""
    if not idList:
      self.log.info("updatePopEntryStatus: no entry to be updated")
      return S_OK()

    sqlStatus = self._escapeString(newStatus)['Value']
    strIdList = [str(iD) for iD in idList]
    sqlIdList = ", ".join(strIdList)
    self.log.verbose("updatePopEntryStatus: sqlIdList %s " % sqlIdList)
    sqlCmd = "UPDATE `Popularity` SET Status=%s WHERE ID IN ( %s )" % (sqlStatus, sqlIdList)
    result = self._update(sqlCmd)
    if not result['OK']:
      return result
    return result

  def insertToDirMetadata(self, directoryDict):
    """Inserts a new entry into the DirMetadata table.

    The input dictionary should contain, for each lfn directory, a
    dictionary with all the necessary metadata provided by the
    bookkeeping
    """

    if not directoryDict:
      return S_OK()
    # get the IDs from the su_Directory table:
    result = self.__getIDs(directoryDict)
    if not result['OK']:
      return result
    dirIDs = result['Value']
    if not dirIDs:
      errMsg = "ERROR! the directory %s was not found in the LFN directory table! " % directoryDict
      self.log.error("%s" % errMsg)
      return S_ERROR(errMsg)
    self.log.verbose("in insertToDirMetadata: dirIDs: %s" % (dirIDs))
    # returns a dictionary of type:
    #  {dir1: ID1, dir2: ID2, ...}
    insertedEntries = 0
    for dirKey in directoryDict:
      if dirKey not in dirIDs:
        self.log.warn("in insertToDirMetadata: found no DID for directory %s" % dirKey)
        continue
      iD = int(dirIDs[dirKey])
      try:
        info = directoryDict[dirKey]
        sqlConfigname = self._escapeString(info['ConfigName'])['Value']
        sqlConfigversion = self._escapeString(info['ConfigVersion'])['Value']
        sqlConditions = self._escapeString(info['ConditionDescription'])['Value']
        sqlProcpass = self._escapeString(info['ProcessingPass'])['Value']
        sqlEvttype = self._escapeString(info['EventType'])['Value']
        sqlFiletype = self._escapeString(info['FileType'])['Value']
        sqlProd = self._escapeString(info['Production'])['Value']
        sqlVisi = self._escapeString(info['Visibility'])['Value']
      except KeyError:
        self.log.error("in insertToDirMetadata: the input dict was not correctly formatted: %s" % directoryDict)
        return S_ERROR("Key error in input dictionary %s" % directoryDict)
      sqlCmd = "INSERT INTO `DirMetadata` ( DID, ConfigName, ConfigVersion, Conditions, ProcessingPass, " \
          "EventType, FileType, Production, Visibility ) VALUES ( %d, %s, %s, %s, %s, %s, %s, %s, %s )" % \
          (iD, sqlConfigname, sqlConfigversion, sqlConditions, sqlProcpass, sqlEvttype, sqlFiletype, sqlProd, sqlVisi)
      self.log.verbose("in insertToDirMetadata, sqlCmd: %s " % sqlCmd)
      result = self._update(sqlCmd)
      if not result['OK']:
        self.log.error("Cannot insert entry: %s" % (result['Message']))
        return S_ERROR(result['Message'])
      insertedEntries += 1
    return S_OK(insertedEntries)

  def getDirMetadata(self, dirList):
    """Return the directory meta-data, which have been previously provided by
    the Bookkeeping and stored in the DirMetadata table.

    The input is a list of LFN directories
    """
    if not dirList:
      return S_OK()
    # get the IDs from the su_Directory table:
    stdDirList = [_standardDirectory(dirPath) for dirPath in dirList]
    result = self.__getIDs(stdDirList)
    if not result['OK']:
      return S_ERROR(result['Message'])
    dirIDs = result['Value']
    if not dirIDs:
      errMsg = "ERROR! the directories %s were not found in the LFN directory table! " % str(stdDirList)
      self.log.error("%s" % errMsg)
      return S_ERROR(errMsg)
    self.log.verbose("in getDirMetadata, dirIDs: %s" % (dirIDs))
    # returns a dictionary of type:
    #  {dir1: ID1, dir2: ID2, ...}
    idList = []
    for dirKey in stdDirList:
      if dirKey not in dirIDs:
        self.log.warn("in getDirMetadata: found no DID for directory %s" % dirKey)
      else:
        idList.append(str(dirIDs[dirKey]))
    sqlCmd = "SELECT DID, ConfigName, ConfigVersion, Conditions, ProcessingPass, EventType, FileType, " \
        "Production, Visibility FROM `DirMetadata` WHERE DID in ( %s )" % ", ".join(["'%s'" % did for did in idList])
    self.log.verbose("in getDirMetadata, sqlCmd: %s " % sqlCmd)
    result = self._query(sqlCmd)
    if not result['OK']:
      return S_ERROR(result['Message'])
    # We must build the list of tuples with the same order than in the input
    metadata = {}
    resList = result['Value']
    for dirPath, dirKey in zip(dirList, stdDirList):
      metadata[dirPath] = ()
      if dirKey in dirIDs:
        did = dirIDs[dirKey]
        for t in resList:
          if t[0] == did:
            metadata[dirPath] = t
            break
    self.log.verbose('getDirMetadata returns %s' % str(metadata))
    return S_OK(metadata)

################################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Gets a list of files from BK and checks if they have common ancestors."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

fixIt = None


def removeFile(lfns, transClient, dm, bkClient):
  """Method for removing a file in the RM as well as in the Transformation
  system."""
  res = bkClient.setFilesInvisible(lfns)
  if res['OK']:
    gLogger.always('Files were made invisible in BK')
  else:
    gLogger.always('Error making files invisible in BK', res['Message'])
  gLogger.always('Removing %d files from disk' % len(lfns))
  res = dm.getReplicas(lfns)
  success = res.get('Value', {}).get('Successful', [])
  failed = res.get('Value', {}).get('Failed', lfns)
  if res['OK'] and success:
    seLfns = {}
    removeFiles = list(success)
    for lfn in success:
      for se in success[lfn]:
        if not se.endswith('ARCHIVE'):
          seLfns.setdefault(se, []).append(lfn)
        else:
          # If there is an archive, we cannot remove the file completely
          if lfn in removeFiles:
            removeFiles.remove(lfn)
    if removeFiles:
      gLogger.always('Removing completely %d files' % len(removeFiles))
      res = dm.removeFile(removeFiles)
      if not res['OK']:
        gLogger.fatal('Error removing %d files' % len(removeFiles), res['Message'])
      else:
        success = res['Value']['Successful']
        failed = res['Value']['Failed']
        gLogger.always('Successfully fully removed %d files that only had the duplicate ancestors' % len(success))
        if failed:
          errors = {}
          for lfn, reason in failed.items():
            errors.setdefault(reason, []).append(lfn)
          gLogger.error(
              'Failed to remove %d files' %
              len(failed), '\n'.join(
                  '%s: %s' %
                  (reason, errors[reason]) for reason in errors))
    success = []
    failed = {}
    for se, replicas in seLfns.items():
      replicas = [lfn for lfn in replicas if lfn not in removeFiles]
      if replicas:
        res = dm.removeReplica(se, replicas)
        if not res['OK']:
          gLogger.fatal('Error removing replicas', res['Message'])
        else:
          success += res['Value']['Successful']
          if res['Value']['Failed']:
            failed[se] = res['Value']['Failed']
    if success:
      gLogger.always('Successfully removed %d replicas of files that only had the duplicate ancestors' % len(success))
    if failed:
      errors = {}
      for se in failed:
        for lfn, reason in failed[se].items():
          errors.setdefault(reason + ' @%s' % se, []).append(lfn)
      gLogger.error('Failed to remove replicas\n', '\n'.join('%s: %s' % (reason, errors[reason]) for reason in errors))
    elif not success and not removeFiles:
      gLogger.always('None of the files had replicas left on disk...')
    if not success:
      return
    res = transClient.getTransformationFiles({'LFN': success})
    if res['OK']:
      transFiles = {}
      processedFiles = {}
      for fileDict in res['Value']:
        if fileDict['Status'] == 'Processed':
          processedFiles.setdefault(fileDict['TransformationID'], []).append(fileDict['LFN'])
        else:
          transFiles.setdefault(fileDict['TransformationID'], []).append(fileDict['LFN'])
      if processedFiles:
        gLogger.always('WARNING: some files were already processed!')
        for transID, lfns in processedFiles.items():
          gLogger.always('%d: %d files' % (transID, len(lfns)))
          gLogger.info('\n'.join(sorted(lfns)))
      for transID, lfns in transFiles.items():
        res = transClient.setFileStatusForTransformation(transID, 'Removed', lfns)
        if res['OK']:
          gLogger.always('%d: %d files set Removed' % (transID, len(lfns)))
          gLogger.info('\n'.join(sorted(lfns)))
        else:
          gLogger.fatal('Error setting %d files to Removed status in transformation %d' % (len(lfns), transID))
  elif failed:
    gLogger.fatal('Failed to get replicas for %d files:\n' % len(failed),
                  res.get('Message', '\n'.join(sorted(set(failed.values
                                                          if isinstance(failed, type({})) else
                                                          [])))))
    gLogger.info('\n'.join(('%s: %s' % (lfn, failed[lfn]) for lfn in sorted(failed))
                           if isinstance(failed, dict)
                           else
                           sorted(failed)))


def analyzeAncestors(commonAncestors, ancestors, transClient, dm, bkClient):
  """Analyse the list of common ancestors and checks whether one can remove
  some files."""
  lfnsToRemove = set()
  allLfns = [lfn for lfnStr in commonAncestors for lfn in lfnStr.split(',')]
  res = bkClient.getFileMetadata(allLfns)
  lfnRuns = {}
  if res['OK']:
    for lfn, metadata in res['Value']['Successful'].items():
      lfnRuns[lfn] = metadata['RunNumber']
  for lfnStr, anc in commonAncestors.items():
    lfns = lfnStr.split(',')
    run = lfnRuns.get(lfns[0], 'Unknown')
    gLogger.always('\n%s (run %s):\n\t%s' % ('\n'.join(lfns), str(run), '\n\t'.join(anc)))
    # Now check if one of the files only has those common ancestors
    for lfn in sorted(lfns, reverse=True):
      if anc == ancestors[lfn]:
        if not fixIt:
          gLogger.always('%s only has those as ancestors. Use --FixIt to remove the file' % lfn)
        else:
          gLogger.always('%s only has those as ancestors. We shall remove the file' % lfn)
          lfnsToRemove.add(lfn)
          lfns.remove(lfn)
          if len(lfns) == 1:
            break
  if lfnsToRemove:
    removeFile(list(lfnsToRemove), transClient=transClient, dm=dm, bkClient=bkClient)


@DIRACScript()
def main():
  global fixIt

  from DIRAC.Core.Base import Script
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  dmScript = DMScript()
  dmScript.registerFileSwitches()
  dmScript.registerBKSwitches()
  depth = 10
  Script.registerSwitch('', 'Depth=', 'Depth to which one should check ancestors (default %d)' % depth)
  Script.registerSwitch('', 'FixIt', 'Remove the files that only have the common ancestors')
  Script.registerSwitch('', 'Verbose', 'Set script level to INFO')

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile]' % Script.scriptName, ]))
  Script.parseCommandLine(ignoreErrors=True)

  fixIt = False
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'Verbose':
      gLogger.setLevel('INFO')
    elif switch[0] == 'FixIt':
      fixIt = True
    elif switch[0] == 'Depth':
      depth = int(switch[1])

  # In case the user asked for specific LFNs
  lfnList = dmScript.getOption('LFNs', [])

  from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
  from DIRAC.DataManagementSystem.Client.DataManager import DataManager
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  dm = DataManager()
  bkClient = BookkeepingClient()
  transClient = TransformationClient()

  cc = ConsistencyChecks(transClient=transClient, dm=dm, bkClient=bkClient)
  cc.bkQuery = dmScript.getBKQuery()
  cc.lfns = lfnList
  cc.ancestorsDepth = depth
  startTime = time.time()
  cc.checkAncestors()

  # Print out the results
  gLogger.always('\nResults (%.1f seconds):' % (time.time() - startTime))

  if not cc.commonAncestors:
    gLogger.always('No files found with common ancestors ==> OK')
  else:
    gLogger.always('Found %d sets of files with common ancestors ==> ERROR' % len(cc.commonAncestors))
    analyzeAncestors(cc.commonAncestors, cc.ancestors, transClient=transClient, dm=dm, bkClient=bkClient)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This script adds a user directory to the DFC or changes a user directory's
ownership."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
from time import time

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  from DIRAC import exit, gLogger

  Script.registerSwitch('', 'User=', '  User name (no default, mandatory)')
  Script.registerSwitch('', 'Recursive', '  Set ownership recursively for existing user')

  Script.parseCommandLine()

  user = None
  recursive = False
  group = 'lhcb_user'
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'User':
      user = switch[1].lower()
    elif switch[0] == 'Recursive':
      recursive = True

  from DIRAC.Core.Security.ProxyInfo import getProxyInfo
  from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown

  res = getProxyInfo()
  if not res['OK']:
    gLogger.fatal("Can't get proxy info", res['Message'])
    exit(1)
  properties = res['Value'].get('groupProperties', [])
  if 'FileCatalogManagement' not in properties:
    gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
    exit(5)

  from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
  dfc = FileCatalogClient()

  if user is None:
    Script.showHelp(exitCode=1)

  exists = False
  initial = user[0]
  baseDir = os.path.join('/lhcb', 'user', initial, user)
  subDirectories = []
  if dfc.isDirectory(baseDir).get('Value', {}).get('Successful', {}).get(baseDir):
    gLogger.always('Directory already existing', baseDir)
    exists = True
    res = dfc.listDirectory(baseDir)
    if res['OK']:
      success = res['Value']['Successful'][baseDir]
      subDirectories = success['SubDirs']
      if success.get('SubDirs') or success.get('Files'):
        gLogger.always(
            'Directory is not empty:', ' %d files / %d subdirectories' %
            (len(success['Files']), len(success['SubDirs']))
        )
      elif recursive:
        gLogger.always("Empty directory, recursive is useless...")
        recursive = False
  if not exists:
    gLogger.always('Creating directory', baseDir)
    res = dfc.createDirectory(baseDir)
    if not res['OK']:
      gLogger.fatal('Error creating directory', res['Message'])
      exit(2)
  else:
    gLogger.always('Change%s ownership of directory' % ' recursively' if recursive else '', baseDir)
  res = chown(baseDir, user, group='lhcb_user', mode=0o755, recursive=False, fcClient=dfc)
  if not res['OK']:
    gLogger.fatal('Error changing directory owner', res['Message'])
    exit(2)
  startTime = time()
  if recursive and subDirectories:
    res = chown(subDirectories, user, group='lhcb_user', mode=0o755, recursive=True, ndirs=1, fcClient=dfc)
    if not res['OK']:
      gLogger.fatal('Error changing directory owner', res['Message'])
      exit(2)
    gLogger.always('Successfully changed owner in %d directories in %.1f seconds' % (res['Value'], time() - startTime))
  else:
    gLogger.always('Successfully changed owner in directory %s in %.1f seconds' % (baseDir, time() - startTime))

  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
  from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeLfnMetadata
  dmScript = DMScript()
  dmScript.setLFNs(baseDir)
  sys.stdout.write('Directory metadata: ')
  sys.stdout.flush()
  executeLfnMetadata(dmScript)
  exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Whatever:

.. code-block:: python

  in_dict = {
    'EventTypeId': 93000000,
    'ConfigVersion': 'Collision10',
    'ProcessingPass': '/Real Data',
    'ConfigName': 'LHCb',
    'ConditionDescription': 'Beam3500GeV-VeloClosed-MagDown',
    'Production':7421
  }
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tempfile
import six

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.DISET.TransferClient import TransferClient

from LHCbDIRAC.BookkeepingSystem.Client import JEncoder
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

__RCSID__ = "$Id$"


@createClient('Bookkeeping/BookkeepingManager')
class BookkeepingClient(Client):
  """This class expose the methods of the Bookkeeping Service."""

  def __init__(self, url=None, **kwargs):
    """c'tor.

    :param str url: can specify a specific URL
    """
    super(BookkeepingClient, self).__init__(**kwargs)
    self.setServer('Bookkeeping/BookkeepingManager')
    if url:
      self.setServer(url)
    self.timeout = 3600
    self.log = gLogger.getSubLogger('BookkeepingClient')

  #############################################################################
  def getAvailableFileTypes(self):
    """It returns all the available files which are registered to the bkk."""
    retVal = self._getRPC().getAvailableFileTypes()
    if retVal['OK']:
      records = []
      parameters = ["FileType", "Description"]
      for record in retVal['Value']:
        records += [list(record)]
      return S_OK({'ParameterNames': parameters, 'Records': records, 'TotalRecords': len(records)})
    return retVal

  #############################################################################
  @staticmethod
  def getFilesWithMetadata(in_dict):
    """It is used for retrieving a files with meta data for a given condition.

    :param dict in_dict: It can contains the following conditions:
      ``ConfigName``, ``ConfigVersion``, ``ConditionDescription``,
      ``EventType``, ``ProcessingPass``,``Production``,``RunNumber``,
      ``FileType``, ``DataQuality``, ``StartDate`` and ``EndDate``
    :return: files with meta data associated
    """
    in_dict = dict(in_dict)
    bkk = TransferClient('Bookkeeping/BookkeepingManager')
    params = JEncoder.dumps(in_dict)
    file_name = tempfile.NamedTemporaryFile()
    retVal = bkk.receiveFile(file_name.name, params)
    if not retVal['OK']:
      return retVal
    value = JEncoder.load(open(file_name.name))
    file_name.close()
    return S_OK(value)

  #############################################################################
  def bulkJobInfo(self, in_dict):
    """It returns the job metadata information for a given condition:

    -a list of lfns
    - a list of DIRAC job ids
    - a list of jobNames

    :param dict in_dict: dictionary which has the following format: in_dict = {'lfn':[],jobId:[],jobName:[]}
    :return: job meta data
    """
    conditions = {}
    if isinstance(in_dict, six.string_types):
      conditions['lfn'] = in_dict.split(';')
    elif isinstance(in_dict, list):
      conditions['lfn'] = in_dict
    else:
      conditions = in_dict

    return self._getRPC().bulkJobInfo(conditions)

  #############################################################################
  def setFileDataQuality(self, lfns, flag):
    """It is used to set the files data quality.

    :param list lfns: list of LFNs or an LFN
    :param str flag: data quality flag: OK, UNCHECKED, etc.
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')

    return self._getRPC().setFileDataQuality(lfns, flag)

  #############################################################################
  def getFileAncestors(self, lfns, depth=0, replica=True):
    """Retrieve file ancestors.

    :param list lfns: list of LFNs
    :param int depth: depth of the processing chane
    :param bool replica: take into account the replica flag.

    :returns: It returns the ancestors of a file with metadata or a list of files
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')

    return self._getRPC().getFileAncestors(lfns, depth, replica)

  #############################################################################
  def getFileDescendants(self, lfns, depth=0, production=0, checkreplica=False):
    """Retrieve the file descendants.

    :param list lfns: list of LFNs
    :param int depth: depth of the processing chane
    :param bool replica: take into account the replica flag.
    :return: descendants of a file or a list of files.
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')

    return self._getRPC().getFileDescendants(lfns, depth, production, checkreplica)

  #############################################################################
  def addFiles(self, lfns):
    """It sets the replica flag Yes for a given list of files.

    :param list lfns: list of LFNs
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().addFiles(lfns)

  #############################################################################
  def removeFiles(self, lfns):
    """It removes the replica flag for a given list of files.

    :param list lfns: list of lfns
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().removeFiles(lfns)

  #############################################################################
  def getFileMetadata(self, lfns):
    """Retrieve the metadata information for a given file or a list of files.

    :param list lfns: list of LFNs
    :return: file metadata
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().getFileMetadata(lfns)

  #############################################################################
  def getFileMetaDataForWeb(self, lfns):
    """This method only used by the web portal. It is same as getFileMetadata.

    :param list lfns: list of LFNs
    :return: file metadata
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().getFileMetaDataForWeb(lfns)

  #############################################################################
  def exists(self, lfns):
    """It used to check the existence of a list of files in the Bookkeeping
    Metadata catalogue.

    :param list lfns: list of LFNs
    :return: a dictionary with the lfns {'lfn':True/False}
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().exists(lfns)

  #############################################################################
  def getRunInformation(self, in_dict):
    """It returns run information and statistics.

    :param dict in_dict: contains a given conditions
    :return: run statistics Number of events, luminosity, etc.
    """
    if 'Fields' not in in_dict:
      in_dict['Fields'] = ['ConfigName', 'ConfigVersion', 'JobStart', 'JobEnd', 'TCK',
                           'FillNumber', 'ProcessingPass', 'ConditionDescription', 'CONDDB', 'DDDB']
    if 'Statistics' in in_dict and not in_dict['Statistics']:
      in_dict['Statistics'] = ['NbOfFiles', 'EventStat', 'FileSize', 'FullStat',
                               'Luminosity', 'InstLumonosity', 'EventType']

    return self._getRPC().getRunInformation(in_dict)

  #############################################################################
  def getRunInformations(self, runnb):
    """It returns run information and statistics.

    :param (int, str) runnb: run number
    :return: run statistics
    """
    # The service expects a int
    try:
      return self._getRPC().getRunInformations(int(runnb))
    except (ValueError, TypeError) as e:
      return S_ERROR("Invalid run number: %s" % repr(e))

#############################################################################
  def getRunFilesDataQuality(self, runs):
    """For retrieve the data quality for runs or files.

    :param list runs: list of run numbers.
    :return: run or file data quality
    """
    if isinstance(runs, six.string_types):
      runs = runs.split(';')
    elif isinstance(runs, six.integer_types):
      runs = [runs]
    return self._getRPC().getRunFilesDataQuality(runs)

  #############################################################################
  def setFilesInvisible(self, lfns):
    """It is used to set the file(s) invisible in the database.

    :paran list lfns: an lfn or list of lfns
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().setFilesInvisible(lfns)

  #############################################################################
  def setFilesVisible(self, lfns):
    """It is used to set the file(s) invisible in the database.

    :param list lfns: an lfn or list of lfns
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().setFilesVisible(lfns)

  #############################################################################
  def getFileTypeVersion(self, lfns):
    """For retrieving the file type version.

    :param list lfns: list of lfns
    :return: file type version
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().getFileTypeVersion(lfns)

  #############################################################################
  def getDirectoryMetadata(self, lfns):
    """For retrieving  meta data information for a given directory.

    :param list lfns: list of lfns
    :return: directory metadata
    """
    if isinstance(lfns, six.string_types):
      lfns = lfns.split(';')
    return self._getRPC().getDirectoryMetadata(lfns)

  #############################################################################
  def getRunsForFill(self, fillid):
    """For retrieving a list of runs.

    :param int fillid: fill number
    :return: runs for a given fill
    """
    try:
      fill = int(fillid)
    except ValueError as ex:
      return S_ERROR(ex)
    return self._getRPC().getRunsForFill(fill)

  #############################################################################
  def deleteSimulationConditions(self, simid):
    """It deletes a given simulation condition.

    :param int simid: simulation condition identifier
    """
    try:
      simid = int(simid)
    except ValueError as ex:
      return S_ERROR(ex)
    return self._getRPC().deleteSimulationConditions(simid)

  #############################################################################
  def getJobInputOutputFiles(self, diracjobids):
    """For retrieving input/output for a given Dirac job.

    :param list diracjobids: list of dirac job ids.
    :return: input/output file(s) for a given dirac job
    """
    if isinstance(diracjobids, six.integer_types):
      diracjobids = [diracjobids]
    return self._getRPC().getJobInputOutputFiles(diracjobids)

  def fixRunLuminosity(self, runnumbers):
    """For fixing the luminosity for a given run(s)

    :param list runnumbers: list of run numbers.
    """
    if isinstance(runnumbers, six.integer_types):
      runnumbers = [runnumbers]
    return self._getRPC().fixRunLuminosity(runnumbers)

  # The following method names are changed in the Bookkeeping client.

  #############################################################################
  def getFiles(self, in_dict):
    """It returns a list of files for a given condition.

    :param dict in_dict: contains a given conditions
    :return: list of files
    """
    in_dict = dict(in_dict)
    bkk = TransferClient('Bookkeeping/BookkeepingManager')
    in_dict['MethodName'] = 'getFiles'
    params = JEncoder.dumps(in_dict)
    file_name = tempfile.NamedTemporaryFile()
    retVal = bkk.receiveFile(file_name.name, params)
    if not retVal['OK']:
      return retVal
    value = JEncoder.load(open(file_name.name))
    file_name.close()
    return value

  def getRunStatus(self, runs):
    """For retrieving the run status.

    :param list runs: list of runs
    :return: run status (finished, or not finished)
    """
    runnumbers = []
    if isinstance(runs, six.string_types):
      runnumbers = [int(run) for run in runs.split(';')]
    elif isinstance(runs, six.integer_types):
      runnumbers += [runs]
    else:
      runnumbers = runs
    return self._getRPC().getRunStatus(runnumbers)

  def getProcessingPass(self, in_dict, path=None):
    """This method is used to recursively browse the processing pass.

    :param dict in_dict: contains a given conditions: ConfigName', 'ConfigVersion', 'ConditionDescription',
      'Production','RunNumber', 'EventType'
    :param str path: To start the browsing you have to define the path as a root: path = '/'
    :return: processing pass for a given conditions. Note: it returns a list with two dictionary. First dictionary
      contains the processing passes while the second dictionary contains the event types.
    """
    if path is None:
      path = '/'
    return self._getRPC().getProcessingPass(in_dict, path)

  def getProductionFilesStatus(self, productionid=None, lfns=None):
    """Status of the files, which belong to a given production.

        :param str/int prodID: production (transformation) ID
        :param list lfns: list of LFNs
        :returns: the file status in the bkk for a given production or a list of lfns.
    """
    if lfns is None:
      lfns = []
    return self._getRPC().getProductionFilesStatus(productionid, lfns)

  def getProductionInformation(self, prodID):
    """ Get the production information.

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with dictionary of production info
    """
    res = self._getRPC().getProductionInformation(prodID)
    if not res['OK']:
      return res

    prodInfo = res['Value']

    res = self.getSteps(prodID)
    if not res['OK']:
      return res

    prodInfo["Steps"] = res['Value']

    return S_OK(prodInfo)

  def getSteps(self, prodID):
    """ Fully resolve the steps of the production in input

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with list of resolved steps
    """
    res = self._getRPC().getSteps(prodID)
    if not res['OK']:
      return res
    steps = res['Value']  # this is an ordered list
    if not steps:
      self.log.error("Production %s does not have recorded steps" % prodID)
      return S_ERROR("No recorded steps")
    # now we check if the first in the list had DDDB and CondDB defined, or not
    # if not, we get the steps of all the previous productions
    if steps[0][4] == steps[0][5] == 'fromPreviousStep':
      # if we are here it is because in the current production none of the steps contain DB tags
      self.log.info("DB tags are not set: they will be retrieved from the parent production(s)",
                    "(prod: %s)" % prodID)
      numberOfSteps = len(steps)
      # Now finding the previous productions
      res = self._getPreviousProductions(prodID)
      if not res['OK']:
        return res
      ancestorProdIDs = res['Value']  # an already ordered list
      if not ancestorProdIDs:
        return S_ERROR("No ancestor productions found")

      for ancestorProdID in ancestorProdIDs:
        res = self._getRPC().getSteps(ancestorProdID)
        if not res['OK']:
          return res
        stepsInAncestorProd = res['Value']
        steps = stepsInAncestorProd + steps
      allResolvedSteps = self._resolveProductionSteps(steps)
      return S_OK(allResolvedSteps[-numberOfSteps:])
    else:
      return S_OK(self._resolveProductionSteps(steps))

  def _resolveProductionSteps(self, steps):
    """ Takes care of resolving the steps of a single production
        (including resolving the DDDB and CondDB tags "fromPreviousStep")

        :param list steps: list of steps (which are tuples)
        :returns: list of resolved steps
    """
    if not steps:
      return []

    # DDDB and CondDB are often registered as "fromPreviousStep", so they should be resolved
    # This will search among the steps in the current production (might not be final)
    # A fair assumption is that dddb and conddb are both either set, or not (so both 'fromPreviousStep').

    productionSteps = [steps[0]]
    for i, nextStep in enumerate(steps[1:]):
      nextStep = list(nextStep)
      if nextStep[4] == 'fromPreviousStep':
        nextStep[4] = productionSteps[i][4]
      if nextStep[5] == 'fromPreviousStep':
        nextStep[5] = productionSteps[i][5]
      productionSteps.append(tuple(nextStep))
    return productionSteps

  def _getPreviousProductions(self, prodID):
    """ Returns an already-ordered list of production(s)
        that were inputs to the provided one

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with list of ancestorProdIDs or S_ERROR
    """
    # Start by getting the RequestID
    res = TransformationClient().getTransformation(prodID, True)
    if not res['OK']:
      self.log.error("Could not retrieve parameters for production",
                     '%d: %s' % (prodID, res['Message']))
      return res
    parameters = res['Value']

    if parameters['Status'] in ('Cleaned', 'Deleted'):
      self.log.notice("The production is Cleaned/Deleted")
      return S_ERROR("The production is Cleaned/Deleted")

    # Now getting the TransformationIDs for the RequestID
    reqID = parameters.get('RequestID')
    if not reqID:
      self.log.error("No RequestID recorded for production", prodID)
      return S_ERROR("No RequestID recorded for production")

    res = ProductionRequestClient().getProductionList(int(reqID))
    if not res['OK']:
      self.log.error("Could not retrieve productions list for request",
                     '%d:%s' % (int(reqID), res['Message']))
      return res
    ancestorProdIDs = res['Value']

    # Now only returning the production IDs prior to prodID (in good order)
    ancestorProdIDs = ancestorProdIDs[0: ancestorProdIDs.index(prodID)]
    ancestorProdIDs.reverse()
    return S_OK(ancestorProdIDs)


class BKClientWithRetry():
  """Utility class wrapping BKClient with retries."""

  def __init__(self, bkClient=None, retries=None):
    if not bkClient:
      bkClient = BookkeepingClient()
    self.bk = bkClient
    self.retries = retries if retries else 5
    self.method = None

  def __getattr__(self, x):
    self.method = x
    return self.__executeMethod

  def __executeMethod(self, *args, **kwargs):
    fcn = getattr(self.bk, self.method)
    for _i in range(self.retries):
      res = fcn(*args, **kwargs)
      if res['OK']:
        break
    return res



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""A mock of the BookkeepingClient, used for testing purposes."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

from mock import MagicMock

bkc_mock = MagicMock()
bkc_mock.sendXMLBookkeepingReport.return_value = {'OK': True, 'Value': ''}
bkc_mock.getFileTypes.return_value = {'OK': True,
                                      'rpcStub': (('Bookkeeping/BookkeepingManager',
                                                   {'skipCACheck': False, 'delegatedGroup': 'diracAdmin',
                                                    'timeout': 3600}), 'getFileTypes', ({}, )),
                                      'Value': {'TotalRecords': 48, 'ParameterNames': ['FileTypes'],
                                                'Records': [['DAVINCIHIST'], ['DIELECTRON.DST'], ['BU2JPSIK.MDST'],
                                                            ['SIM'], ['BD2JPSIKS.MDST'],
                                                            ['BU2JPSIK.DST'], ['BUBDBSSELECTION.DST'],
                                                            ['LAMBDA.DST'], ['BSMUMUBLIND.DST'], ['HADRONIC.DST']]}}
bkc_mock.getFileMetadata.return_value = {
    'OK': True,
    'Value': {
        'Successful': {
            'foo': {
                'ADLER32': None,
                'FileType': 'SDST',
                'FullStat': None,
                'GotReplica': 'Yes',
                'DataqualityFlag': 'OK',
                'EventStat': 101,
                'EventType': 12365501,
                'FileSize': 1000,
                'GUID': '7EA19086-1E36-E611-9457-001E67398520',
                'RunNumber': 93718},
            'bar': {
                'ADLER32': None,
                'FileType': 'SDST',
                'FullStat': None,
                'GotReplica': 'Yes',
                'DataqualityFlag': 'OK',
                'EventStat': 102,
                'EventType': 12365501,
                'FileSize': 2000,
                'GUID': '7EA19086-1E36-E611-9457-001E67398520',
                'RunNumber': 93720}}},
    'rpcStub': (
        ('Bookkeeping/BookkeepingManager',
         ))}
bkc_mock.getFileTypeVersion.return_value = {'OK': True,
                                            'Value': {'lfn1': 'ROOT',
                                                      'lfn2': 'MDF'}}
bkc_mock.getFileDescendants.return_value = {'OK': True,
                                            'Value': {'Failed': [],
                                                      'NotProcessed': [],
                                                      'Successful': {'aa.raw': ['bb.raw', 'bb.log']},
                                                      'WithMetadata': {'aa.raw': {'bb.raw': {'FileType': 'RAW',
                                                                                             'RunNumber': 97019,
                                                                                             'GotReplica': 'Yes'},
                                                                                  'bb.log': {'FileType': 'LOG',
                                                                                             'GotReplica': 'Yes'}
                                                                                  }
                                                                       }
                                                      }
                                            }


class BookkeepingClientFake(object):
  """ a fake BookkeepingClient - replicating some of the methods
  """

  def getAvailableSteps(self, stepID):

    if stepID == {'StepId': 123}:
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId', 'StepName', 'ApplicationName', 'ApplicationVersion',
                                           'OptionFiles', 'Visible', 'ExtraPackages', 'ProcessingPass', 'OptionsFormat',
                                           'DDDB', 'CONDDB', 'DQTag', 'SystemConfig'],
                        'Records': [[123, 'Stripping14-Stripping', 'DaVinci', 'v2r2',
                                     'optsFiles', 'Yes', 'eps', 'procPass', '',
                                     '', '123456', '', '']]}}
    elif stepID in ({'StepId': 456}, {'StepId': 987}):
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId', 'StepName', 'ApplicationName', 'ApplicationVersion',
                                           'OptionFiles', 'Visible', 'ExtraPackages', 'ProcessingPass', 'OptionsFormat',
                                           'DDDB', 'CONDDB', 'DQTag', 'SystemConfig'],
                        'Records': [[456, 'Merge', 'LHCb', 'v1r2',
                                     'optsFiles', 'Yes', 'eps', 'procPass', '',
                                     '', 'fromPreviousStep', '', 'x86_64-slc6-gcc49-opt']]}}
    elif stepID == {'StepId': 789}:
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId', 'StepName', 'ApplicationName', 'ApplicationVersion',
                                           'OptionFiles', 'Visible', 'ExtraPackages', 'ProcessingPass', 'OptionsFormat',
                                           'DDDB', 'CONDDB', 'DQTag', 'SystemConfig'],
                        'Records': [[789, 'MergeHisto', 'Noether', 'v1r2',
                                     'optsFiles', 'Yes', 'eps', 'procPass', '',
                                     '', 'fromPreviousStep', '', 'x86_64-slc6-gcc49-opt']]}}
    elif stepID == {'StepId': 125080}:
      options125080 = '$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;'
      options125080 += '$DECFILESROOT/options/11102400.py;'
      options125080 += '$LBPYTHIA8ROOT/options/Pythia8.py;'
      options125080 += '$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;'
      options125080 += '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py'
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'ExtraPackages',
                  'ProcessingPass',
                  'Visible',
                  'Usable',
                  'DDDB',
                  'CONDDB',
                  'DQTag',
                  'OptionsFormat',
                  'OptionFiles',
                  'isMulticore',
                  'SystemConfig',
                  'mcTCK',
                  'ExtraOptions'],
              'Records': [
                  [
                      125080,
                      'Sim08a',
                      'Gauss',
                      'v45r3',
                      'AppConfig.v3r171',
                      'Sim08a',
                      'Y',
                      'Yes',
                      'Sim08-20130503-1',
                      'Sim08-20130503-1-vc-mu100',
                      '',
                      '',
                      options125080,
                      'N',
                      'x86_64-slc5-gcc43-opt',
                      '',
                      '']]}}
    elif stepID == {'StepId': 124620}:
      options124620 = '$APPCONFIGOPTS/Boole/Default.py;'
      options124620 += '$APPCONFIGOPTS/Boole/DataType-2012.py;'
      options124620 += '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py'
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'ExtraPackages',
                  'ProcessingPass',
                  'Visible',
                  'Usable',
                  'DDDB',
                  'CONDDB',
                  'DQTag',
                  'OptionsFormat',
                  'OptionFiles',
                  'isMulticore',
                  'SystemConfig',
                  'mcTCK',
                  'ExtraOptions'],
              'Records': [
                  [
                      124620,
                      'Digi13',
                      'Boole',
                      'v26r3',
                      'AppConfig.v3r164',
                      'Digi13',
                      'N',
                      'Yes',
                      'Sim08-20130503-1',
                      'Sim08-20130503-1-vc-mu100',
                      '',
                      '',
                      options124620,
                      'N',
                      'x86_64-slc5-gcc43-opt',
                      '',
                      '']]}}
    elif stepID == {'StepId': 124621}:
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'ExtraPackages',
                  'ProcessingPass',
                  'Visible',
                  'Usable',
                  'DDDB',
                  'CONDDB',
                  'DQTag',
                  'OptionsFormat',
                  'OptionFiles',
                  'isMulticore',
                  'SystemConfig',
                  'mcTCK',
                  'ExtraOptions'],
              'Records': [
                  [
                      124621,
                      'Digi13',
                      'Brunel',
                      'v26r3',
                      'AppConfig.v3r164',
                      'Digi13',
                      'N',
                      'Yes',
                      'Sim08-20130503-1',
                      'Sim08-20130503-1-vc-mu100',
                      '',
                      '',
                      'someBrunelOptions',
                      'N',
                      'x86_64-slc5-gcc43-opt',
                      '',
                      '']]}}
    elif stepID == {'StepId': 999}:
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId',
                                           'OptionFiles',
                                           'ExtraOptions',
                                           'SystemConfig',
                                           'mcTCK',
                                           'isMulticore',
                                           'StepName',
                                           'ApplicationName',
                                           'ApplicationVersion',
                                           'Visible',
                                           'Usable',
                                           'ProcessingPass',
                                           'ExtraPackages',
                                           'DDDB',
                                           'CONDDB',
                                           'DQTag'],
                        'Records': [[999,
                                     '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py',
                                     '',
                                     'x86_64-slc5-gcc49-opt',
                                     '',
                                     'N',
                                     'Stripping28',
                                     'DaVinci',
                                     'v41r3',
                                     'Yes',
                                     'Yes',
                                     'Stripping28',
                                     'AppConfig.v3r306',
                                     'dddb-20150724',
                                     'cond-20161011',
                                     '']]}}
    elif stepID == {'StepId': 998}:
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId',
                                           'OptionFiles',
                                           'ExtraOptions',
                                           'SystemConfig',
                                           'mcTCK',
                                           'isMulticore',
                                           'StepName',
                                           'ApplicationName',
                                           'ApplicationVersion',
                                           'Visible',
                                           'Usable',
                                           'ProcessingPass',
                                           'ExtraPackages',
                                           'DDDB',
                                           'CONDDB',
                                           'DQTag'],
                        'Records': [[998,
                                     '',
                                     '',
                                     'x86_64-slc6-gcc49-opt',
                                     '',
                                     'N',
                                     'Stripping28',
                                     'DaVinci',
                                     'v41r3',
                                     'Yes',
                                     'Yes',
                                     'Stripping28',
                                     'AppConfig.v3r306',
                                     'dddb-20150724',
                                     'cond-20161011',
                                     '']]}}
    elif stepID == {'StepId': 997}:
      return {'OK': True,
              'Value': {'TotalRecords': 1,
                        'ParameterNames': ['StepId',
                                           'OptionFiles',
                                           'ExtraOptions',
                                           'SystemConfig',
                                           'mcTCK',
                                           'isMulticore',
                                           'StepName',
                                           'ApplicationName',
                                           'ApplicationVersion',
                                           'Visible',
                                           'Usable',
                                           'ProcessingPass',
                                           'ExtraPackages',
                                           'DDDB',
                                           'CONDDB',
                                           'DQTag'],
                        'Records': [[997,
                                     '$APPCONFIGOPTS/Brunel/DataType-2016.py;$APPCONFIGOPTS/Brunel/PbPb-GECs.py',
                                     '',
                                     'x86_64-slc6-gcc49-opt',
                                     '',
                                     'N',
                                     'Stripping28',
                                     'DaVinci',
                                     'v41r3',
                                     'Yes',
                                     'Yes',
                                     'Stripping28',
                                     'AppConfig.v3r306',
                                     'dddb-20150724',
                                     'cond-20161011',
                                     '']]}}
    elif stepID == {'StepId': 996}:
      options996 = '$APPCONFIGOPTS/Brunel/DataType-2016.py;'
      options996 += '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;'
      options996 += '$APPCONFIGOPTS/Brunel/PbPb-GECs.py'
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'OptionFiles',
                  'ExtraOptions',
                  'SystemConfig',
                  'mcTCK',
                  'isMulticore',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'Visible',
                  'Usable',
                  'ProcessingPass',
                  'ExtraPackages',
                  'DDDB',
                  'CONDDB',
                  'DQTag'],
              'Records': [
                  [
                      996,
                      options996,
                      '',
                      'x86_64-slc6-gcc49-opt',
                      '',
                      'N',
                      'Stripping28',
                      'DaVinci',
                      'v41r3',
                      'Yes',
                      'Yes',
                      'Stripping28',
                      'AppConfig.v3r306',
                      'dddb-20150724',
                      'cond-20161011',
                      '']]}}
    elif stepID == {'StepId': 1098}:
      options1098 = '$APPCONFIGOPTS/Brunel/DataType-2016.py;'
      options1098 += '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;'
      options1098 += '$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py'
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'OptionFiles',
                  'ExtraOptions',
                  'SystemConfig',
                  'mcTCK',
                  'isMulticore',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'Visible',
                  'Usable',
                  'ProcessingPass',
                  'ExtraPackages',
                  'DDDB',
                  'CONDDB',
                  'DQTag'],
              'Records': [
                  [
                      1098,
                      options1098,
                      '',
                      'x86_64-slc6-gcc49-opt',
                      '',
                      'N',
                      'Stripping28',
                      'DaVinci',
                      'v41r3',
                      'Yes',
                      'Yes',
                      'Stripping28',
                      'AppConfig.v3r306',
                      'dddb-20150724',
                      'cond-20161011',
                      '']]}}
    elif stepID == {'StepId': 1099}:
      options1099 = '$APPCONFIGOPTS/Brunel/DataType-2016.py;'
      options1099 += '$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;'
      options1099 += '$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py'
      return {
          'OK': True,
          'Value': {
              'TotalRecords': 1,
              'ParameterNames': [
                  'StepId',
                  'OptionFiles',
                  'ExtraOptions',
                  'SystemConfig',
                  'mcTCK',
                  'isMulticore',
                  'StepName',
                  'ApplicationName',
                  'ApplicationVersion',
                  'Visible',
                  'Usable',
                  'ProcessingPass',
                  'ExtraPackages',
                  'DDDB',
                  'CONDDB',
                  'DQTag'],
              'Records': [
                  [
                      1099,
                      options1099,
                      '',
                      'x86_64-slc6-gcc49-opt',
                      '',
                      'N',
                      'Stripping28',
                      'DaVinci',
                      'v41r3',
                      'Yes',
                      'Yes',
                      'Stripping28',
                      'AppConfig.v3r306',
                      'dddb-20150724',
                      'cond-20161011',
                      '']]}}

  def getStepInputFiles(self, stepID):
    if stepID == 123:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['SDST', 'Y']]}}

    if stepID == 456:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['BHADRON.DST', 'Y'], ['CALIBRATION.DST', 'Y']]}}
    if stepID == 789:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['DAVINCIHIST', 'Y'], ['BRUNELHIST', 'Y']]}}
    if stepID == 125080:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['', 'Y']]}}
    if stepID == 124620 or stepID == 124621:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['SIM', 'N']]}}
    if stepID == 999 or stepID == 998 or stepID == 997 or stepID == 996 or stepID == 1098 or stepID == 1099:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['SDST', 'Y']]}}

  def getStepOutputFiles(self, stepID):
    if stepID == 123:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['BHADRON.DST', 'Y'], ['CALIBRATION.DST', 'Y']]}}
    if stepID == 456:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['BHADRON.DST', 'Y'], ['CALIBRATION.DST', 'Y']]}}
    if stepID == 789:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['ROOT', 'Y']]}}
    if stepID == 125080:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['SIM', 'Y']]}}
    if stepID == 124620:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['DIGI', 'N']]}}
    if stepID == 124621:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['MDF', 'Y']]}}
    if stepID == 999 or stepID == 998 or stepID == 997 or stepID == 996 or stepID == 1098 or stepID == 1099:
      return {'OK': True,
              'Value': {'TotalRecords': 7,
                        'ParameterNames': ['FileType', 'Visible'],
                        'Records': [['BHADRON.DST', 'Y'], ['CALIBRATION.DST', 'Y']]}}


# Some steps definitions, for testing purposes

options125080 = '$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;'
options125080 += '$DECFILESROOT/options/11102400.py;'
options125080 += '$LBPYTHIA8ROOT/options/Pythia8.py;'
options125080 += '$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;'
step125080 = {'StepId': 125080, 'StepName': 'Sim08a',
              'ApplicationName': 'Gauss', 'ApplicationVersion': 'v45r3', 'ExtraOptions': '',
              'OptionFiles': options125080,
              'Usable': 'Yes', 'Visible': 'Y', 'ExtraPackages': 'AppConfig.v3r171',
              'ProcessingPass': 'Sim08a', 'OptionsFormat': '',
              'prodStepID': "125080['']", 'SystemConfig': 'x86_64-slc5-gcc43-opt',
              'DDDB': 'Sim08-20130503-1', 'CONDDB': 'Sim08-20130503-1-vc-mu100', 'DQTag': '', 'isMulticore': 'N',
              'mcTCK': '',
              'visibilityFlag': [{'Visible': 'N', 'FileType': 'SIM'}],
              'fileTypesIn': [''],
              'fileTypesOut': ['SIM']}

stepMC = {'StepId': 123, 'StepName': 'Stripping14-Stripping',
          'ApplicationName': 'Gauss', 'ApplicationVersion': 'v2r2', 'ExtraOptions': '',
          'OptionFiles': 'optsFiles', 'Visible': 'Yes', 'ExtraPackages': 'eps',
          'ProcessingPass': 'procPass', 'OptionsFormat': '',
          'prodStepID': "123['SDST']", 'SystemConfig': '',
          'DDDB': '', 'CONDDB': '123456', 'DQTag': '', 'isMulticore': 'N',
          'mcTCK': '',
          'visibilityFlag': [{'Visible': 'N', 'FileType': 'SIM'}],
          'fileTypesIn': [],
          'fileTypesOut': ['SIM']}

stepMC2 = {'StepId': 123, 'StepName': 'Stripping14-Stripping',
           'ApplicationName': 'DaVinci', 'ApplicationVersion': 'v2r2', 'ExtraOptions': '',
           'OptionFiles': 'optsFiles', 'Visible': 'Yes', 'ExtraPackages': 'eps',
           'ProcessingPass': 'procPass', 'OptionsFormat': '',
           'prodStepID': "123['SDST']", 'SystemConfig': '',
           'DDDB': '', 'CONDDB': '123456', 'DQTag': '', 'isMulticore': 'N',
           'mcTCK': '',
           'visibilityFlag': [{'Visible': 'Y', 'FileType': 'DST'}],
           'fileTypesIn': ['SIM'],
           'fileTypesOut': ['DST']}

step1Dict = {'StepId': 123, 'StepName': 'Stripping14-Stripping',
             'ApplicationName': 'DaVinci', 'ApplicationVersion': 'v2r2', 'ExtraOptions': '',
             'OptionFiles': 'optsFiles', 'Visible': 'Yes', 'ExtraPackages': 'eps',
             'ProcessingPass': 'procPass', 'OptionsFormat': '',
             'prodStepID': "123['SDST']", 'SystemConfig': '',
             'DDDB': '', 'CONDDB': '123456', 'DQTag': '', 'isMulticore': 'N',
             'mcTCK': '',
             'visibilityFlag': [{'Visible': 'N', 'FileType': 'BHADRON.DST'},
                                {'Visible': 'N', 'FileType': 'CALIBRATION.DST'}],
             'fileTypesIn': ['SDST'],
             'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST']}

step2Dict = {'StepId': 456, 'StepName': 'Merge',
             'ApplicationName': 'LHCb', 'ApplicationVersion': 'v1r2', 'ExtraOptions': '',
             'OptionFiles': 'optsFiles', 'Visible': 'Yes', 'ExtraPackages': 'eps',
             'ProcessingPass': 'procPass', 'OptionsFormat': '', 'SystemConfig': 'x86_64-slc6-gcc49-opt',
             'prodStepID': "456['BHADRON.DST', 'CALIBRATION.DST']",
             'DDDB': '', 'CONDDB': '123456', 'DQTag': '', 'isMulticore': 'N',
             'mcTCK': '',
             'visibilityFlag': [{'Visible': 'N', 'FileType': 'BHADRON.DST'},
                                {'Visible': 'N', 'FileType': 'CALIBRATION.DST'}],
             'fileTypesIn': ['BHADRON.DST', 'CALIBRATION.DST'],
             'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST']}

step124620Dict = {'StepId': 124620, 'StepName': 'Digi13',
                  'ApplicationName': 'Boole', 'ApplicationVersion': 'v26r3', 'ExtraOptions': '',
                  'OptionFiles': 'optsFiles', 'Visible': 'N', 'ExtraPackages': 'AppConfig.v3r164',
                  'ProcessingPass': 'Digi13', 'OptionsFormat': '',
                  'prodStepID': "124620['SIM']", 'SystemConfig': 'x86_64-slc5-gcc43-opt',
                  'DDDB': 'Sim08-20130503-1', 'CONDDB': 'Sim08-20130503-1-vc-mu100', 'DQTag': '', 'isMulticore': 'N',
                  'mcTCK': '', 'Usable': 'Yes',
                  'visibilityFlag': [{'Visible': 'N', 'FileType': 'DIGI'}],
                  'fileTypesIn': ['SIM'],
                  'fileTypesOut': ['DIGI']}

step124621Dict = {'StepId': 124621, 'StepName': 'Digi13',
                  'ApplicationName': 'Brunel', 'ApplicationVersion': 'v26r3', 'ExtraOptions': '',
                  'OptionFiles': 'optsFiles', 'Visible': 'N', 'ExtraPackages': 'AppConfig.v3r164',
                  'ProcessingPass': 'Digi13', 'OptionsFormat': '',
                  'prodStepID': "124621['SIM']", 'SystemConfig': 'x86_64-slc5-gcc43-opt',
                  'DDDB': 'Sim08-20130503-1', 'CONDDB': 'Sim08-20130503-1-vc-mu100', 'DQTag': '', 'isMulticore': 'N',
                  'mcTCK': '', 'Usable': 'Yes',
                  'visibilityFlag': [{'Visible': 'Y', 'FileType': 'MDF'}],
                  'fileTypesIn': ['SIM'],
                  'fileTypesOut': ['MDF']}

stepStripp = {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 123, 'ApplicationVersion': 'v28r3p1',
              'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'SystemConfig': '',
              'ProcessingPass': 'Merging', 'Visible': 'N', 'DDDB': 'head-20110302', 'mcTCK': '',
              'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py', 'CONDDB': 'head-20110407',
              'visibilityFlag': [],
              'fileTypesIn': ['SDST'],
              'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST', 'CHARM.MDST', 'CHARMCOMPLETEEVENT.DST']}
mergeStep = {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 456, 'ApplicationVersion': 'v28r3p1',
             'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'SystemConfig': '',
             'ProcessingPass': 'Merging', 'Visible': 'N', 'DDDB': 'head-20110302', 'mcTCK': '',
             'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py', 'CONDDB': 'head-20110407',
             'visibilityFlag': [],
             'fileTypesIn': ['BHADRON.DST', 'CALIBRATION.DST', 'PID.MDST'],
             'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST', 'PID.MDST']}
mergeStepBHADRON = {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 456, 'ApplicationVersion': 'v28r3p1',
                    'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'SystemConfig': '',
                    'prodStepID': "456['BHADRON.DST']",
                    'ProcessingPass': 'Merging', 'Visible': 'N', 'DDDB': 'head-20110302', 'mcTCK': '',
                    'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py', 'CONDDB': 'head-20110407',
                    'visibilityFlag': [],
                    'fileTypesIn': ['BHADRON.DST'],
                    'fileTypesOut': ['BHADRON.DST']}
mergeStepCALIBRA = {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 456, 'ApplicationVersion': 'v28r3p1',
                    'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'SystemConfig': '',
                    'prodStepID': "456['CALIBRATION.DST']",
                    'ProcessingPass': 'Merging', 'Visible': 'N', 'DDDB': 'head-20110302', 'mcTCK': '',
                    'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py', 'CONDDB': 'head-20110407',
                    'visibilityFlag': [],
                    'fileTypesIn': ['CALIBRATION.DST'],
                    'fileTypesOut': ['CALIBRATION.DST']}
mergeStepPIDMDST = {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 456, 'ApplicationVersion': 'v28r3p1',
                    'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'SystemConfig': '',
                    'prodStepID': "456['PID.MDST']",
                    'ProcessingPass': 'Merging', 'Visible': 'N', 'DDDB': 'head-20110302', 'mcTCK': '',
                    'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py', 'CONDDB': 'head-20110407',
                    'visibilityFlag': [],
                    'fileTypesIn': ['PID.MDST'],
                    'fileTypesOut': ['PID.MDST']}

stepHistoMergingDict = {'StepId': 789, 'StepName': 'MergeHisto',
                        'ApplicationName': 'Noether', 'ApplicationVersion': 'v1r2', 'ExtraOptions': '',
                        'OptionFiles': 'optsFiles', 'Visible': 'Yes', 'ExtraPackages': 'eps',
                        'ProcessingPass': 'procPass', 'OptionsFormat': '', 'SystemConfig': 'x86_64-slc6-gcc49-opt',
                        'prodStepID': "789['DAVINCIHIST', 'BRUNELHIST']",
                        'DDDB': '', 'CONDDB': '123456', 'DQTag': '', 'isMulticore': 'N',
                        'mcTCK': '',
                        'visibilityFlag': [{'Visible': 'Y', 'FileType': 'ROOT'}],
                        'fileTypesIn': ['DAVINCIHIST', 'BRUNELHIST'],
                        'fileTypesOut': ['ROOT']}



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: UserStorageQuotaAgent

.. module: UserStorageQuotaAgent

:synopsis: UserStorageQuotaAgent obtains the usage by each user from the StorageUsageDB
  and compares with a quota present in the CS.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gConfig, S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

__RCSID__ = "$Id$"

AGENT_NAME = 'DataManagement/UserStorageQuotaAgent'


class UserStorageQuotaAgent(AgentModule):
  """.. class:: UserStorageQuotaAgent.

  :param int deafultQuota: default quota in MB
  :param NotificationClient notificationClient: NotificationClient instance
  :param LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB.StorageUsageDB storageUsageDB:
         StorageUsageDB or RPC client pointing to StorageUsageDB
  """

  defaultQuota = 1000
  notificationClient = None
  storageUsageDB = None

  def __init__(self, *args, **kwargs):
    """c'tor."""
    AgentModule.__init__(self, *args, **kwargs)

    self.notificationClient = NotificationClient()
    try:
      self.storageUsageDB = StorageUsageDB()
    except SystemExit:
      self.storageUsageDB = StorageUsageClient()

    self.defaultQuota = gConfig.getValue('/Registry/DefaultStorageQuota', self.defaultQuota)  # Default is 1TB

  def initialize(self):
    """agent initialisation.

    :param self: self reference
    """
    # This sets the Default Proxy to used as that defined under
    # /Operations/Shifter/DataManager
    # the shifterProxy option in the Configuration can be used to change this default.
    self.am_setOption('shifterProxy', 'DataManager')

    self.log.info("initialize: Default quota found to be %d GB" % self.defaultQuota)
    return S_OK()

  def execute(self):
    """execution of one cycle.

    :param self: self reference
    """
    res = self.storageUsageDB.getUserStorageUsage()
    usageDict = res['Value']

    byteToGB = 1000 * 1000 * 1000.0

    managerMsg = ""
    errorMsg = ""
    self.log.info("Determining quota usage for %s users." % len(usageDict))
    for userName in sorted(usageDict):
      usageGB = usageDict[userName] / byteToGB
      res = gConfig.getOptionsDict('/Registry/Users/%s' % userName)
      if not res['OK']:
        msg = "Username not found in the CS: %s using %.2f GB" % (userName, usageGB)
        errorMsg += msg + '\n'
        self.log.error(msg)
        continue
      elif 'Email' not in res['Value']:
        msg = "CS does not contain email information for user %s" % userName
        errorMsg += msg + '\n'
        self.log.error(msg)
        continue
      elif "Quota" not in res['Value']:
        userQuota = float(self.defaultQuota)
      else:
        userQuota = float(res['Value']['Quota'])
      userMail = res['Value']['Email']
      # Different behaviour for 90% exceeded, 110% exceeded and 150% exceeded
      msg = None
      if (1.5 * userQuota) < usageGB:
        msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (userName, (usageGB * 100) / userQuota,
                                                           '%', userQuota, usageGB)
        self.log.info(msg)
        self.sendBlockedMail(userName, userMail, userQuota, usageGB)
        self.log.info("!!!!!!!!!!!!!!!!!!!!!!!!REMEMBER TO MODIFY THE ACLs and STATUS HERE!!!!!!!!!!!!!!!!!")
      elif (1.0 * userQuota) < usageGB:
        msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (userName, (usageGB * 100) / userQuota,
                                                           '%', userQuota, usageGB)
        self.log.info(msg)
        self.sendSecondWarningMail(userName, userMail, userQuota, usageGB)
      elif (0.9 * userQuota) < usageGB:
        msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (userName, (usageGB * 100) / userQuota,
                                                           '%', userQuota, usageGB)
        self.log.info(msg)
        self.sendFirstWarningMail(userName, userMail, userQuota, usageGB)
      if msg:
        managerMsg += msg + "\n"
    if managerMsg or errorMsg:
      if managerMsg:
        managerMsg = (
            "Mails have been sent to the following list of users " +
            "being close to or above quota:\n\n" +
            managerMsg
        )
      if errorMsg:
        managerMsg += "\nThe following errors have been found by the UserStorageQuotaAgent:\n" + errorMsg
      fromAddress = 'LHCb Data Manager <lhcb-datamanagement@cern.ch>'
      toAddress = 'lhcb-datamanagement@cern.ch'
      self.notificationClient.sendMail(toAddress, "User quota warnings", managerMsg, fromAddress)
    return S_OK()

  def sendFirstWarningMail(self, userName, userMail, quota, usage):
    """first warning email.

    :param self: self reference
    :param str userName: DIRAC user name
    :param str userMail: email address
    :param int quota: default quota
    :param float usage: space currently used
    """
    msgbody = """
This mail has been generated automatically.

You have received this mail because you are approaching your Grid storage usage quota of %s GB.

You are currently using %.1f GB.

Please reduce you usage by removing some files. If you have reduced your usage in the last 24 hours
please ignore this message.

Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""" % (int(quota), usage)
    fromAddress = 'LHCb Data Manager <lhcb-datamanagement@cern.ch>'
    subject = 'Grid storage use near quota (%s)' % userName
    toAddress = userMail
    self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)

  def sendSecondWarningMail(self, userName, userMail, quota, usage):
    """second warning email.

    :param self: self reference
    :param str userName: DIRAC user name
    :param str userMail: email address
    :param int quota: default quota
    :param float usage: space currently used
    """
    msgbody = """
This mail has been generated automatically.

You have received this mail because your Grid storage usage has exceeded your quota of %sGB.

You are currently using %.1f GB.

Please reduce you usage by removing some files. If you have reduced your usage in the last 24 hours
please ignore this message.

Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""" % (int(quota), usage)
    fromAddress = 'LHCb Data Manager <lhcb-datamanagement@cern.ch>'
    subject = 'Grid storage use over quota (%s)' % userName
    toAddress = userMail
    self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)

  def sendBlockedMail(self, userName, userMail, quota, usage):
    """send blocked email.

    :param self: self reference
    :param str userName: DIRAC user name
    :param str userMail: email adress
    :param int quota: default quota
    :param float usage: space used
    """
    msgbody = """
This mail has been generated automatically.

You have received this mail because your Grid storage usage has exceeded your quota of %s GB.

You are currently using %.1f GB.

Your account could soon been given a lower priority and your jobs will run at a lower pace if you
don't create space.
If you have reduced your usage in the last 24 hours please ignore this message.
Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""" % (int(quota), usage)

    fromAddress = 'LHCb Data Manager <lhcb-datamanagement@cern.ch>'
    subject = 'Grid storage use blocked (%s)' % userName
    toAddress = userMail
    self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Returns metadata of the job step(s) that created a (list of) LFNs,
or all steps for a (list of) JobID
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  from DIRAC import gLogger
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, printDMResult

  bkScript = DMScript()
  bkScript.registerFileSwitches()
  Script.registerSwitch('', 'Summary', '   Only report job IDs creating the LFN(s)')
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... [LFN|File|JobID]' % Script.scriptName,
      'Arguments:',
      '  LFN:      Logical File Name',
      '  File:     Name of the file with a list of LFNs',
      '  JobID:    DIRAC job identifier']))

  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()
  jobIDList = []
  for arg in args:
    if arg.isdigit():
      jobIDList.append(int(arg))
    else:
      bkScript.setLFNsFromFile(arg)
  lfnList = bkScript.getOption('LFNs', [])
  if not lfnList and not jobIDList:
    Script.showHelp(exitCode=1)
  summary = False
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'Summary':
      summary = True

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  if lfnList:
    retVal = BookkeepingClient().bulkJobInfo(lfnList)
  elif jobIDList:
    retVal = BookkeepingClient().bulkJobInfo({'jobId': jobIDList})
    summary = False
  jobLFNs = {}
  if retVal['OK']:
    success = retVal['Value']['Successful']
    jobDict = {}
    for item in list(success):
      jobs = success.pop(item)
      # Note that item may be a jobID or an LFN
      if not item.isdigit():
        #   item is an LFN, hence only one job step (job in BK)
        lfn = item
        job = jobs[0]
        jobLFNs.setdefault(job['Name'], []).append(lfn)
        jobDict.setdefault(job['Name'], job)
      else:
        jobID = item
        # item is a jobID, hence we may have a list of job steps
        if int(jobID) in retVal['Value']['Failed']:
          # For some reason sometimes (always?) the jobID is also in Failed
          retVal['Value']['Failed'].remove(int(jobID))
        jobStr = 'Job %s' % jobID
        stepDict = {'Job':
                    {
                        'CPUTIME': 0.0,
                        'ExecTime': 0.0,
                        'NumberOfSteps': len(jobs)}
                    }
        for job in jobs:
          # Group common info into a separate dictionary (and remove it from steps)
          job.pop('DIRACJobId')
          # ... and yes, it is TotalLumonosity ;-)
          for info in ('DIRACVersion', 'Location', 'Production', 'TotalLumonosity',
                       'WNCACHE', 'WNCPUHS06', 'WNCPUPOWER', 'WNMEMORY', 'WNMJFHS06', 'WNMODEL', 'WORKERNODE'):
            stepDict['Job'][info] = job.pop(info, None)
          # Sum up CPU and wall clock times
          for info in ('CPUTIME', 'ExecTime'):
            stepDict['Job'][info] += job[info]
          step = 'Step %s' % job.pop('Name')
          stepDict[step] = job
        success[jobStr] = stepDict

    # When LFNs were given, rearrange the success dictionary in case several files were created by the same step
    if not summary and jobLFNs:
      # Group files produced by the same job
      lfnsByJob = {}
      for name, job in jobDict.items():  # Can be an iterator
        # For each jobID get set of LFNs and job steps
        lfnsByJob.setdefault(job['DIRACJobId'], []).append((jobLFNs[name], job))
      for jobID, lfnJobs in lfnsByJob.items():  # Can be an iterator
        jobStr = 'Job %s' % jobID
        # Split job and step information
        stepDict = {jobStr:
                    {}
                    }
        lfnSet = set()
        for lfns, job in lfnJobs:
          job.pop('DIRACJobId')
          lfnSet.update(lfns)
          for info in ('DIRACVersion', 'Location', 'Production', 'TotalLumonosity',
                       'WNCACHE', 'WNCPUHS06', 'WNCPUPOWER', 'WNMEMORY', 'WNMJFHS06', 'WNMODEL', 'WORKERNODE'):
            stepDict[jobStr][info] = job.pop(info, None)
          job[' LFN'] = ','.join(sorted(lfns))
          stepDict['Step %s' % job.pop('Name')] = job
        success[','.join(sorted(lfnSet))] = stepDict

  if summary and jobLFNs:
    gLogger.always('List of DIRAC jobs:')
    gLogger.always(','.join(sorted(set('%s' % val['DIRACJobId'] for val in jobDict.values()))))  # Can be an iterator
  else:
    printDMResult(retVal, empty="File/job does not exist in the Bookkeeping")


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client plug-in for the RAWIntegrity catalog. This exposes a single method to
add files to the RAW IntegrityDB.

USED at OnLine
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase

__RCSID__ = "$Id$"


class RAWIntegrityClient(FileCatalogClientBase):

  # List of common File Catalog methods implemented by this client
  WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + ['addFile']

  def __init__(self, url='', **kwargs):

    self.serverURL = 'DataManagement/RAWIntegrity' if not url else url
    super(RAWIntegrityClient, self).__init__(self.serverURL, **kwargs)
    self.rawIntegritySrv = self._getRPC()

  def isOK(self):
    """Returns valid."""
    return self.valid

  @checkCatalogArguments
  def addFile(self, lfns):
    failed = {}
    successful = {}
    for lfn, info in lfns.items():
      pfn = str(info['PFN'])
      size = int(info['Size'])
      se = str(info['SE'])
      guid = str(info['GUID'])
      checksum = str(info['Checksum'])
      res = self.rawIntegritySrv.addFile(lfn, pfn, size, se, guid, checksum)
#       rpc = self._getRPC()
#       rpc.addFile( lfn, pfn, size, se, guid, checksum )
      if not res['OK']:
        failed[lfn] = res['Message']
      else:
        successful[lfn] = True

    resDict = {'Failed': failed,
               'Successful': successful}
    return S_OK(resDict)
